/*
 * USB Driver for WebCams using the Ali M5602 and M5603c chipset.
 * 
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation, version 2.
 *
 */

#ifndef M560X_H_
#define M560X_H_

/* Driver info */

#define DRIVER_AUTHOR "Ali m560x Linux Driver Project <m560x@x3ng.com.br>"
#define DRIVER_DESC "Ali M560x Webcam Driver"

/* TODO this value must be inserted into linux/videodev.h */
#define VID_HARDWARE_M560X        40

enum camera_type {
	CAMERA_M5602 = 0,
	CAMERA_M5603 = 1,
};

#define M560X_BYTES_PER_PIXEL   3
#define M560X_MAX_READ_SIZE     (512*242+128)
#define M560X_MAX_FRAME_SIZE    (M560X_BYTES_PER_PIXEL*320*240)
#define M560X_FRAMES            2
#define M560X_MAX_FRAMES        32
#define M560X_HEADER_SIZE       64

#define M560X_FIRMWARE_FILENAME "m560x.bin"
#define FIRMWARE_STEP_SIZE      0x3e8
#define M5602_SCRIPT_FILENAME   "m5602.csr"
#define M5603_SCRIPT_FILENAME   "m5603.csr"
/* number of simultaneous URB submitted to data endpoint (Windows drv : 20 */
#define NB_ISOC_URBS 0x8
/* number of ISO packet per URB (Windows driver uses 10) */

//#define PKTSIZE 0x3ff
//#define PKTSIZE 0x800
#define ISOC_PKTSIZE 0xc00
//#define PKTNB 19
#define NB_ISOC_PACKETS         20
#define M560X_NUMFRAMES         2
#define M560X_URB_TIMEOUT       msecs_to_jiffies(2 * NB_ISOC_PACKETS)

#include "m560x_sensor.h"

struct nlist{
  struct nlist *next;
  char *name;
  unsigned short entry_value;
};

enum {
  STATE_SCANNING,   /* Scanning for start */
  STATE_HEADER,     /* Parsing header */
  STATE_LINES,      /* Parsing lines */
};

enum m560x_frame_state {
  FRAME_UNUSED,     /* Unused (no MCAPTURE) */
  FRAME_QUEUED,     /* Ready to start grabbing */
  FRAME_GRABBING,   /* In the process of being grabbed into */
  FRAME_DONE,       /* Finished grabbing, but not been synced yet */
  FRAME_ERROR,      /* Something bad happened while processing */
};

enum m560x_io_method {
  IO_NONE,
  IO_READ,
  IO_MMAP,
};

enum m560x_stream_state {
  STREAM_OFF,
  STREAM_INTERRUPT,
  STREAM_ON,
};

enum m560x_dev_state {
  DEV_INITIALIZED = 0x01,
  DEV_DISCONNECTED = 0x02,
  DEV_MISCONFIGURED = 0x04,
};

struct m560x_frame_t {
  int framenum;
  unsigned char * bufmem;
  unsigned char * rawdata;
  int depth;                    /* bytes per pixel */
  int width;                    /* width of frame sent from camera */
  int height;                   /* height of frame sent from camera */
  int bytes_recvd;              /* Number of bytes received from camera */
  long bytes_read;              /* Amount that has been read() */
  volatile int grabstate;
  int scanstate;
  wait_queue_head_t wq;         /* Processes waiting */
  struct v4l2_buffer buf;
  enum m560x_frame_state state;
  struct list_head frame;
  unsigned long vma_use_count;
};


static DECLARE_MUTEX(m560x_sysfs_lock);
static DECLARE_RWSEM(m560x_disconnect);

/* Structure to hold all of our device specific stuff */
struct m560x_camera {
  char *name;
  enum camera_type camera_type;
  struct m560x_sensor * sensor;
  
  struct semaphore cam_lock;            /* guard against simultaneous 
                                         * accesses to the camera */
  unsigned int VideoMode;               /* 0x0100 = VGA, 
                                         * 0x0000 = CIF, 
                                         * 0x0300 = QVGA */
  int SupportedModes;
  int CIF;
  int VGA;
  int QVGA;
  int cwidth;             /* camera width */
  int cheight;            /* camera height */
  int maxwidth;           /* max video width */
  int maxheight;          /* max video height */
  int vwidth;             /* current width for video window */
  int vheight;            /* current height for video window */
 
  int is_initialized;
  
  struct device dev;
  struct video_device *vdev;        /* the v4l video device for this device */
  
  struct usb_device *udev;          /* the usb device for this camera */
  struct usb_interface *interface;  /* the interface for this usb device */
  unsigned char *control_buffer;
  unsigned char *bayer_frame;
  struct nlist **hashtab;

  unsigned char *int_in_buffer; /* the buffer to receive INT data */
  size_t int_in_size;           /* the size of the INT receive buffer */
  __u8 int_in_endpointAddr;     /* the address of the INT in endpoint */

  struct urb *urb[NB_ISOC_URBS];
  unsigned char *isoc_in_buffer[NB_ISOC_URBS];   /* the buffer to receive
                                                 * ISOC data */
//  unsigned char       * isoc_in_buffer;    /* the buffer to receive ISOC data */
  size_t        isoc_in_size;          /* the size of the ISOC receive buffer */
  __u8          isoc_in_endpointAddr;  /* the address of the ISOC in endpoint */

  int           user;             /* user count for exclusive use */
  int           streaming;        /* Are we streaming video? */
  unsigned int  palette;
  struct semaphore buf_lock;
  unsigned char *framebuffers;
  unsigned char *databuffers;
  unsigned char *datacollector;
  int curframe;                   /* Current receiving sbuf */
  int grabframe;
  unsigned char frame_index;
  struct m560x_frame_t *frame_current, frame[M560X_MAX_FRAMES];
  int packetCount;
  int readCount;
  int frameCount;
  int error;
  int dropped;
  u32 frame_count, nbuffers, nreadbuffers;
  struct semaphore lock;           /* to lock the structure */
  wait_queue_head_t wq;            /* Processes waiting */
  struct kref kref;
  struct list_head inqueue,outqueue;
  enum m560x_io_method io;
  enum m560x_stream_state stream;
  enum m560x_dev_state state;
  u8 users;
 
  struct semaphore dev_sem, fileop_sem;
  spinlock_t queue_lock;
  wait_queue_head_t open, wait_frame, wait_stream;
};

/*****************************************************************************/
#define to_m560x_camera(d) container_of(d, struct m560x_camera, kref)


#define DBG_TRACE  0x10
#define DBG_SCRIPT 0x20
#define DBG_DATA 0x40

#define PDEBUG(level, fmt, args...) \
do { \
  if (debug & level)     \
    info("[%s:%d] " fmt, __FUNCTION__, __LINE__ , ## args); \
} while (0)


#endif /*M560X_H_*/
