/*
 * Ali M560x driver
 *
 * Copyright (C) 2005 m560x Linux Driver Project <m560x@x3ng.com.br>
 * Code to usb interface and ALI driver sofwtare:
 * copyright (c) 2006 Willem Duinker
 * v4l2 interface modelled after:
 * V4L2 driver for SN9C10x PC Camera Controllers
 * Copyright (C) 2004-2005 by Luca Risolia <luca.risolia@studio.unibo.it>
 * for potential future code sharing
 *
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License as
 *      published by the Free Software Foundation, version 2.
 *
 * This driver is based on the 2.6.3 version of drivers/usb/usb-skeleton.c
 * but has been rewritten to be easy to read and use, as no locks are now
 * needed anymore.
 *
 */

#include <linux/autoconf.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/vmalloc.h>
#include <linux/kref.h>
#include <linux/version.h>
#include <asm/uaccess.h>
#include <linux/videodev.h>
#include <linux/usb.h>
#include <linux/firmware.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 16)
 #include <media/v4l2-dev.h>
#endif

#define VIDIOC_S_CTRL_OLD       _IOW  ('V', 28, struct v4l2_control)
#define DRIVER_VERSION "0.4.0"
#define DRIVER_VERSION_NUMERICAL 0x400
#include "m560x.h"
#define M560X_FIRMWARE_FILENAME "m560x.bin"
#define M560X_INF_FILENAME "m560x.inf"
#define FIRMWARE_STEP_SIZE 0x3e8
#define HASHSIZE 199

/* module parameter */
static unsigned int debug = 0;

/* table of usb-devices that work with this driver */
static struct usb_device_id m560x_table [] = {
	/* m5602 */
	{ USB_DEVICE(0x0402, 0x5602), .driver_info=CAMERA_M5602 },
	/* m5603 */
	{ USB_DEVICE(0x0402, 0x5603), .driver_info=CAMERA_M5603 },
	{ USB_DEVICE(0x041e, 0x4038), .driver_info=CAMERA_M5603 },
	{ }                                     /* Terminating entry */
};
MODULE_DEVICE_TABLE(usb, m560x_table);

M5603_SENSOR_TABLE
M5602_SENSOR_TABLE


/***************** buffer memory management ***********************************/

static void *rvmalloc(size_t size)
{
	void *mem;
	unsigned long adr;

	size = PAGE_ALIGN(size);
	mem = vmalloc_32(size);
	if (!mem)
		return NULL;

	memset(mem, 0, size); /* Clear the ram out, no junk to the user */
	adr = (unsigned long) mem;
	while (size > 0) {
		SetPageReserved(vmalloc_to_page((void *)adr));
		adr += PAGE_SIZE;
		size -= PAGE_SIZE;
	}

	return mem;
}

static void rvfree(void *mem, size_t size)
{
	unsigned long adr;

	if (!mem)
		return;

	adr = (unsigned long) mem;
	while ((long) size > 0) {
		ClearPageReserved(vmalloc_to_page((void *)adr));
		adr += PAGE_SIZE;
		size -= PAGE_SIZE;
	}
	vfree(mem);
}

static u32 m560x_request_buffers(struct m560x_camera * cam, u32 count,
                                 enum m560x_io_method io)
{
/*  struct v4l2_pix_format *p = &(cam->sensor->pix_format);
  struct v4l2_rect *r = &(cam->sensor->cropcap.bounds);
  const size_t imagesize = cam->module_param.force_munmap ||
      io == IO_READ ?
                           (p->width * p->height * p->priv) / 8 :
      (r->width * r->height * p->priv) / 8;*/
  const size_t imagesize = 640 * 480;
  const size_t buffersize = 0x50000;
  void *buff = NULL;
  void *raw = NULL;
  u32 i;

//   cam->datacollector = rvmalloc(PAGE_ALIGN(700*484));

  if (count > M560X_MAX_FRAMES)
    count = M560X_MAX_FRAMES;

  cam->nbuffers = count;
  while (cam->nbuffers > 0) {
    raw = rvmalloc(cam->nbuffers * PAGE_ALIGN(buffersize));
    if ((buff = rvmalloc(cam->nbuffers * PAGE_ALIGN(3*imagesize))))
      break;
    cam->nbuffers--;
  }

  for (i = 0; i < cam->nbuffers; i++) {
    cam->frame[i].rawdata = raw + i*PAGE_ALIGN(buffersize);
    cam->frame[i].bufmem = buff + i*PAGE_ALIGN(3*imagesize);
    cam->frame[i].buf.index = i;
    cam->frame[i].buf.m.offset = i*PAGE_ALIGN(3*imagesize);
    //cam->frame[i].buf.length = 3 * imagesize;
    cam->frame[i].buf.length = 3 * 640 * 480;
    cam->frame[i].buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    cam->frame[i].buf.sequence = 0;
    cam->frame[i].buf.field = V4L2_FIELD_NONE;
    cam->frame[i].buf.memory = V4L2_MEMORY_MMAP;
    cam->frame[i].buf.flags = 0;
  }
  PDEBUG(DBG_TRACE, "m560x_request_buffers %d", cam->nbuffers);
  return cam->nbuffers;
}


static void m560x_release_buffers(struct m560x_camera * cam)
{
  PDEBUG(DBG_TRACE, "buffers %d", cam->nbuffers);
  if (cam->nbuffers) {
    rvfree(cam->frame[0].rawdata,
           cam->nbuffers * PAGE_ALIGN(0x50000));
    rvfree(cam->frame[0].bufmem,
           cam->nbuffers * PAGE_ALIGN(cam->frame[0].buf.length));
    cam->nbuffers = 0;
  }

  PDEBUG(DBG_TRACE, "buffers released");
}


static void m560x_empty_framequeues(struct m560x_camera * cam)
{
  u32 i;

  INIT_LIST_HEAD(&cam->inqueue);
  INIT_LIST_HEAD(&cam->outqueue);

  for (i = 0; i < M560X_MAX_FRAMES; i++) {
    cam->frame[i].state = FRAME_UNUSED;
    cam->frame[i].buf.bytesused = 0;
  }
}


static void m560x_queue_unusedframes(struct m560x_camera * cam)
{
  unsigned long lock_flags;
  u32 i;
  PDEBUG(DBG_TRACE, "m560x_queue_unusedframes %d", cam->nbuffers);
  for (i = 0; i < cam->nbuffers; i++) {
    if (cam->frame[i].state == FRAME_UNUSED) {
      cam->frame[i].state = FRAME_QUEUED;
      spin_lock_irqsave(&cam->queue_lock, lock_flags);
      list_add_tail(&cam->frame[i].frame, &cam->inqueue);
      spin_unlock_irqrestore(&cam->queue_lock, lock_flags);
    }
  }
}


/*********** ALI USB interface ************************************************/

unsigned char message_1c[]= {
  0x23, 0x74, 0x81, 0x06,
  0x23, 0x62, 0x81, 0x80,
  0x13, 0x6a, 0x81, 0x00,
  0x13, 0x6b, 0x81, 0x00,
  0x13, 0x6c, 0x81, 0x00,
  0x13, 0x6c, 0x81, 0x00,
  0x13, 0x68, 0x81, 0x12,
};

int read_m5602_register(struct m560x_camera *cam, unsigned short ali_register)
{
  struct usb_device *udev = cam->udev;
  u8 *buff = cam->control_buffer;
  int result = 0;

  result = usb_control_msg(udev, usb_rcvctrlpipe(udev, 0),
                                0x04, 0xc0, 0x0014, ali_register, buff,
                                1, 0x100);
  PDEBUG(DBG_SCRIPT, "register %04x result %02x value %02x",
                                ali_register, result, *buff);
  return (result >= 0) ? (int)(*buff) : -1;
}

int write_m5602_register(struct m560x_camera *cam, unsigned short ali_register,
                         unsigned char value)
{
  struct usb_device *udev = cam->udev;
  u8 *buff = cam->control_buffer;
  int result = 0;
  union ali_reg {
    unsigned short value;
    unsigned char u[2];
  } reg;

  reg.value = ali_register;
  buff[0] = 0x13;
  buff[1] = reg.u[0];
  buff[2] = reg.u[1];
  buff[3] = value;

  result = usb_control_msg(udev, usb_sndctrlpipe(udev, 0),
                                0x04, 0x40, 0x0019, 0x0000, buff, 4, 0x100);
  PDEBUG(DBG_SCRIPT, "%02x %2x %02x \t\t\t; write_04 %d\t",
                                buff[1], buff[2], value, result);
  return (result >= 0) ? (int)(*buff) : -1;
}

int read_m5602_sensor_register(struct m560x_camera *cam,
                               unsigned char register_address,
                               unsigned char register_id)
{
  u8 *buff = cam->control_buffer;
  int result = 0;
  unsigned char a_value;
  unsigned char a_value2;

  /* should test on busy bit */
  do {
    result = read_m5602_register(cam, 0x816d);
  } while ((result & 0x80) == 0x80);
  write_m5602_register(cam, 0x816a, register_address);
  write_m5602_register(cam, 0x816b, register_id);
  write_m5602_register(cam, 0x8168, 0x1a);
  a_value = read_m5602_register(cam, 0x816c);
  a_value2 = read_m5602_register(cam, 0x816c);
  buff[0] = a_value;
  buff[1] = a_value2;
  PDEBUG(DBG_SCRIPT, "read m5602 register %02x %02x result %02x %02x",
                                register_address,
                                register_id, buff[0], buff[1]);
  return result;
}

int write_m5602_sensor_register(struct m560x_camera *cam,
                                unsigned char register_address,
                                unsigned char register_id,
                                unsigned char value,
                                unsigned char value2)
{
  struct usb_device *udev = cam->udev;
  u8 *buff = cam->control_buffer;
  int result = 0;
  unsigned char a_value = value;
  unsigned char a_value2 = value2;

  memcpy(buff, message_1c, 0x1c);
  buff[11] = register_address;
  buff[15] = register_id;
  buff[19] = a_value;
  buff[23] = a_value2;

  result = usb_control_msg(udev, usb_sndctrlpipe(udev, 0),
                                0x04, 0x40, 0x0019, 0x0000, buff, 0x1c, 0x100);
  PDEBUG(DBG_SCRIPT, "%02x %02x %02x %02x\t; write_1c %d\t",
                                register_address, register_id,
                                a_value, a_value2, result);
  return (result >= 0) ? (int)(*buff) : -1;
}


int or_m5602_sensor_register(struct m560x_camera *cam,
                             unsigned char register_address,
                             unsigned char register_id,
                             unsigned char value,
                             unsigned char value2)
{
  u8 *buff = cam->control_buffer;
  int result = 0;
  unsigned char a_value ;
  unsigned char a_value2 ;

  result = read_m5602_sensor_register(cam, register_address, register_id);
  PDEBUG(DBG_SCRIPT, "or mask %02x %02x", value, value2);
  PDEBUG(DBG_SCRIPT, "or in %02x %02x", buff[0], buff[1]);
  a_value = buff[0] | value;
  a_value2 = buff[1] | value2;
  PDEBUG(DBG_SCRIPT, "or out %02x %02x", a_value, a_value2);
  result = write_m5602_sensor_register(cam, register_address, register_id,
                                                a_value, a_value2);

  return 0;
}

int and_m5602_sensor_register(struct m560x_camera *cam,
                              unsigned char register_address,
                              unsigned char register_id,
                              unsigned char value,
                              unsigned char value2)
{
  u8 *buff = cam->control_buffer;
  int result = 0;
  unsigned char a_value ;
  unsigned char a_value2 ;

  result = read_m5602_sensor_register(cam, register_address, register_id);
  PDEBUG(DBG_SCRIPT, "and mask %02x %02x", value, value2);
  PDEBUG(DBG_SCRIPT, "and in %02x %02x", buff[0], buff[1]);
  a_value = buff[0] & value;
  a_value2 = buff[1] & value2;
  PDEBUG(DBG_SCRIPT, "and out %02x %02x", a_value, a_value2);
  result = write_m5602_sensor_register(cam, register_address, register_id,
                                                a_value, a_value2);

  return 0;
}

static int getTokens(int *tokens, char *line, int max) {
  int i = 0;
  int count = 0;
  int in_string = 0;

  do {
    if (line[i] == '\n') {
      break;
    }
    else if (line[i] == '\t' || line[i] == ' ' ||
             line[i] == ','  || line[i] == ';' ||
             line[i] == '\r') {
      in_string = 0;
      line[i] = '\0';
    }
    else if (in_string == 0) {
      in_string = 1;
      tokens[count++] = i;
    }

    i++;
  } while (count < max);

  return count;
}

/* Returns pointer to the beginning of the next line in buf, or NULL if
 * end of buf is reached without finding a newline.
 */
unsigned char *script_next_line(u8 *buf, int len)
{
	const u8 *beg = buf;
	buf = memscan(buf, '\n', len);
	if(buf == beg + len || *buf != '\n')
		return NULL;
	buf++;
	return buf;
}

/* Finds section of fw beginning with a line that starts with "[section]".
 * Returns pointer to the beginning of the section header line, or NULL if 
 * section is not found (or out of memory).
 */
unsigned char *script_find_section(u8 *fw, int fw_len, const char *section)
{
	const u8 *end = fw + fw_len;
	int sec_len = strlen(section);
	u8 *sec = kmalloc(sec_len + 2, GFP_KERNEL);

	if(!sec)
		return NULL;

	sec[0] = '[';
	memcpy(sec + 1, section, sec_len);
	sec[sec_len + 1] = ']';

	while (fw && fw < end) {
		if(!memcmp(fw, sec, sec_len))
			break;
		fw = script_next_line(fw, end-fw);
	}

	kfree(sec);

	return fw;
}

int m5602_init_from_script(struct m560x_camera *cam, char * pitem)
{
  const struct firmware *firmware;
  u8 *fw_end;
  u8 *line;
  int retval = 0;
  int result = 0;
  int tokens[20];
  u8 type_id;
  u8 value, value1;
  u16 reg_value;
  u8 reg_id, reg_address;

  PDEBUG(DBG_SCRIPT, "m5602 init from script %s", pitem);
  retval = request_firmware(&firmware, M5602_SCRIPT_FILENAME, &cam->udev->dev);
  if(retval) {
    PDEBUG(DBG_SCRIPT, "request_firmware failed: " M5602_SCRIPT_FILENAME );
    return retval;
  }

  fw_end = firmware->data + firmware->size;

  line = script_find_section(firmware->data, firmware->size, pitem);
  if(!line) {
    retval = -EINVAL;
    goto fail;
  }

  // process appropriate part of script
  while ((line = script_next_line(line, fw_end - line))) {
    type_id = line[0];
    if(type_id == '[')
    	break;
    switch (type_id) {
    case 'S':
      getTokens(tokens, line, 3);
      reg_value = (unsigned short)simple_strtoul(line + tokens[1], NULL, 0);
      value = (unsigned char)simple_strtoul(line + tokens[2], NULL, 0);
      PDEBUG(DBG_SCRIPT, "%c %04x %02x\t", type_id, reg_value, value);
      switch (type_id) {
        case 'S':
          PDEBUG(DBG_SCRIPT, "register write %04x %02x\t", reg_value, value);
          write_m5602_register(cam, reg_value, value);
          break;
      }
      break;
    case 'W':
    case 'O':
    case 'A':
    case 'V':
      getTokens(tokens, line, 5);
      PDEBUG(DBG_SCRIPT, "tokens %s %s %s %s %s\n",
                       line + tokens[0],
                       line + tokens[1],
                       line + tokens[2],
                       line + tokens[3],
                       line + tokens[4]);
      reg_address = (unsigned char)simple_strtoul(line + tokens[1], NULL, 0);
      reg_id = (unsigned char)simple_strtoul(line + tokens[2], NULL, 0);
      value =  (unsigned char)simple_strtoul(line + tokens[3], NULL, 0);  // for kernel simple_strtoul
      value1 = (unsigned char)simple_strtoul(line + tokens[4], NULL, 0);  // for kernel simple_strtoul
      PDEBUG(DBG_SCRIPT, "command %c %02x %02x %02x %02x\n",
                              type_id, reg_address, reg_id, value, value1);
      switch (type_id) {
      case 'W':
        write_m5602_sensor_register(cam, reg_address, reg_id, value,
                                                            value1);
        break;
      case 'V':
        read_m5602_sensor_register(cam, reg_address, reg_id);
        break;
      case 'O':
        or_m5602_sensor_register(cam, reg_address, reg_id, value, value1);
        break;
      case 'A':
        and_m5602_sensor_register(cam, reg_address, reg_id, value, value1);
        break;
      }
      break;
    default:
      break;

    }
  }

fail:
  release_firmware(firmware);
  return result;
}


static int read_ali_register(struct m560x_camera *cam, __u16 ali_register) {
  struct usb_device *udev = cam->udev;
  u8 *buff = cam->control_buffer;
  int result = 0;

  result = usb_control_msg(udev, usb_rcvctrlpipe(udev, 0),
                                0x04, 0xc0, 0x14, ali_register, buff,
				1, 0x100);
  PDEBUG(DBG_SCRIPT, "register %04x result %02x value %02x",
                                ali_register, result, *buff);
  return (result >= 0) ? (int)(*buff) : -1;
}

int write_ali_register(struct m560x_camera *cam, __u16 ali_register,
                       unsigned char value)
{
  struct usb_device *udev = cam->udev;
  u8 *buff = cam->control_buffer;
  int result = 0;

  buff[0] = value;

  result = usb_control_msg(udev, usb_sndctrlpipe(udev, 0),
                                0x04, 0x40, 0x13, ali_register, buff,
				1, 100);
  PDEBUG(DBG_SCRIPT, "register %04x result %02x value %02x",
                                ali_register, result, *buff);
  return (result >= 0) ? (int)(*buff) : -1;
}

int or_ali_register(struct m560x_camera *cam, __u16 ali_register,
                    unsigned char mask)
{
  struct usb_device *udev = cam->udev;
  u8 *buff = cam->control_buffer;
  int result;

  result = usb_control_msg(udev, usb_rcvctrlpipe(udev, 0),
                                0x04, 0xc0, 0x0014, ali_register, buff,
                                1, 0x100);
  PDEBUG(DBG_SCRIPT, "register %04x result %02x value %02x",
                                ali_register, result, *buff);
  if (result != 1 ) {
    return -1;
  }
  buff[0] |= mask;
  PDEBUG(DBG_SCRIPT, "mask %02x data to return %02x", mask, buff[0]);
  result = usb_control_msg(udev, usb_sndctrlpipe(udev, 0),
                                0x04, 0x40, 0x0013, ali_register, buff,
                                1, 100);
  PDEBUG(DBG_SCRIPT, "register %04x result %02x value %02x",
                                ali_register, result, *buff);
  return (result >= 0) ? (int)(*buff) : -1;
}

int and_ali_register(struct m560x_camera *cam, __u16 ali_register, __u8 mask)
{
  struct usb_device *udev = cam->udev;
  u8 *buff = cam->control_buffer;
  int result;

  result = usb_control_msg(udev, usb_rcvctrlpipe(udev, 0),
                                0x04, 0xc0, 0x0014, ali_register, buff,
                                1, 0x100);
  PDEBUG(DBG_SCRIPT, "register %04x result %02x value %02x",
                                ali_register, result, *buff);
  if (result != 1 ) {
    return -1;
  }
  buff[0] &= mask;
  PDEBUG(DBG_SCRIPT, "mask %02x data to return %02x", mask, buff[0]);
  result = usb_control_msg(udev, usb_sndctrlpipe(udev, 0),
                                0x04, 0x40, 0x0013, ali_register, buff,
                                1, 100);
  PDEBUG(DBG_SCRIPT, "register %04x result %02x value %02x",
                                ali_register, result, *buff);
  return (result >= 0) ? (int)(*buff) : -1;
}


int read_sensor_register(struct m560x_camera *cam, __u16 ali_register,
                         unsigned char * i2c_data)
{
//   struct m560x_sensor *s = cam->sensor;
  struct usb_device *udev = cam->udev;
   u8 *buff = cam->control_buffer;
   int result = 0;

   result = usb_control_msg(udev, usb_rcvctrlpipe(udev, 0),
                                0x04, 0xc0, 0x0018, ali_register, buff,
                                1, 100);
   PDEBUG(DBG_TRACE, "read register %04x result %02x value %02x",
                                ali_register, result, *buff);
   i2c_data[0] = 1;
   i2c_data[1] = buff[0];

   return result;
}

int write_sensor_register(struct m560x_camera *cam, __u16 ali_register,
                          unsigned char * i2c_data)
{
  struct m560x_sensor *s = cam->sensor;
  struct usb_device *udev = cam->udev;
  u8 *buff = cam->control_buffer;
  int result = 0;
  int i = 0;
  unsigned char register_id;

  for (i = 0; i < i2c_data[0]; i++) {
    buff[i] = i2c_data[i+1];
  }

  result = usb_control_msg(udev, usb_sndctrlpipe(udev, 0),
                                0x04, 0x40, 0x17, ali_register, buff,
                                2, 100);
  PDEBUG(DBG_SCRIPT, "write register %04x result %02x value %02x %02x",
                                ali_register, result, buff[0], buff[1]);

  register_id = (unsigned char) ((ali_register & 0xff00) >> 8);
  s->i2c_write(cam, register_id, i2c_data);
  return (result >= 0) ? (int)(*buff) : -1;
}

int halt_dsp(struct m560x_camera *cam) {
  u8 *buff = cam->control_buffer;
  int result = 0;

  result = read_ali_register(cam, 0x8000);
  PDEBUG(DBG_SCRIPT, "0x8000 %02x", result);
  result = write_ali_register(cam, 0x8000, 0x40);
  PDEBUG(DBG_SCRIPT, "0x8000 %02x", result);
  result = read_ali_register(cam, 0x8140);
  PDEBUG(DBG_SCRIPT, "0x8140 %02x", result);

  return (result >= 0) ? (int)(*buff) : -1;
}

int restart_dsp(struct m560x_camera *cam)
{
  struct usb_device *udev = cam->udev;
  u8 *buff = cam->control_buffer;
  int result = 0;

  buff[0] = 0x00;
  result = usb_control_msg(udev, usb_sndctrlpipe(udev, 0),
                                0x04, 0x40, 0x10, 0x0000, buff, 1, 100);
  PDEBUG(DBG_SCRIPT, "register %04x result %02x value %02x",
                                0x0000, result, *buff);
  return (result >= 0) ? (int)(*buff) : -1;
}


static unsigned hash(char *s)
{
  unsigned hashval;

  for (hashval = 0; * s!= '\0'; s++)
    hashval = *s + 31 * hashval;
  return hashval % HASHSIZE;
}


static struct nlist *lookup(char *s, struct nlist **hashtab)
{
  struct nlist *np;

  for (np = hashtab[hash(s)];np != NULL;np = np->next) {
    if (strcmp(s, np->name) == 0)
      return np;
  }
  return NULL;
}


int m5603_init_from_script(struct m560x_camera *cam, char *pitem)
{
  const struct firmware *firmware;
  u8 *fw_end;
  u8 *line;
  int retval;
  u8 type_id;
  int result = 0;
  int tokens[20];
  struct nlist * np;
  u8 value, value2;
  u16 reg_value;
  u8 i2c_data[3];
  u8 slave_id;

  PDEBUG(DBG_SCRIPT, "pitem = %s", pitem);
  retval = request_firmware(&firmware, M5603_SCRIPT_FILENAME, &cam->udev->dev);
  if(retval) {
    PDEBUG(DBG_SCRIPT, "request_firmware failed: " M5603_SCRIPT_FILENAME);
    return retval;
  }

  fw_end = firmware->data + firmware->size;

  line = script_find_section(firmware->data, firmware->size, pitem);
  if(!line)
  	return -EINVAL; 

  // process appropriate part of script
  while ((line = script_next_line(line, fw_end - line))) {
    type_id = line[0];
    if(type_id == '[')
    	break;
    switch (type_id) {
    case 'R':
    case 'O':
    case 'A':
      getTokens(tokens, line, 3);
      np = lookup(line + tokens[1], cam->hashtab);
      if (np == NULL )
        return -1;
      value = (unsigned char) simple_strtoul(line + tokens[2], NULL, 0);
      switch (type_id) {
      case 'R':
        write_ali_register(cam, np->entry_value, value);
        break;
      case 'O':
        or_ali_register(cam, np->entry_value, value);
        break;
      case 'A':
        and_ali_register(cam, np->entry_value, value);
        break;
      case 'S':
        read_ali_register(cam, np->entry_value);
        break;
      default:
        break;
      }
      break;
    case 'I':
      getTokens(tokens, line, 5);
      slave_id = simple_strtoul(line + tokens[1], NULL, 0);
      reg_value = (unsigned int) (
                        (simple_strtoul(line + tokens[2], NULL, 0) << 8) |
                        (simple_strtoul(line + tokens[1], NULL, 0)));
      value = (unsigned char) simple_strtoul(line + tokens[3], NULL, 0);  // for kernel simple_strtoul
      value2 = (unsigned char) simple_strtoul(line + tokens[4], NULL, 0);  // for kernel simple_strtoul
      PDEBUG(DBG_SCRIPT, "%c %04x %02x %02x\t", type_id, reg_value,
                                                value, value2);
      if (slave_id == 0xba) {
        i2c_data[0] = 2;
        i2c_data[1] = (unsigned char)simple_strtoul(line + tokens[3], NULL, 0);
        i2c_data[2] = (unsigned char)simple_strtoul(line + tokens[4], NULL, 0);
      }
      write_sensor_register(cam, reg_value, i2c_data);
      break;
    default:
      break;
    }
  }

  release_firmware(firmware);
  return result;
}

int load_firmware(struct m560x_camera *cam)
{
  const struct firmware *firmware;
  int retval;
  unsigned int offset = 0;
  int count = 0;
  int size, i;
  unsigned char *buff, *fwbuf, *pbuf, *rbuf;

  struct usb_device *udev = cam->udev;

  buff = kmalloc(0x40, GFP_KERNEL);
  retval = request_firmware(&firmware, M560X_FIRMWARE_FILENAME,
                             &cam->udev->dev);
  if (retval) {
    err("request_firmware() for " M560X_FIRMWARE_FILENAME " failed\n");
    return retval;
  }
  PDEBUG(DBG_SCRIPT, "size of firmware file %#zx\n", firmware->size);

  size = firmware->size;
  fwbuf = kmalloc(size, GFP_KERNEL);
  for (i = 0; i < size; i++ ) {
    fwbuf[i] = firmware->data[i];
    if (i < 5)
      PDEBUG(DBG_SCRIPT, "test result %02x\n", fwbuf[i]);
  }
  pbuf = fwbuf;

  while ((size -= FIRMWARE_STEP_SIZE) > 0) {
    count = FIRMWARE_STEP_SIZE;
    PDEBUG(DBG_SCRIPT, "do max count %04x %d\n", offset, count);
    retval = usb_control_msg(cam->udev, usb_sndctrlpipe(udev, 0),
                                0x04, 0x40, 0x15, offset, pbuf, count, 100);
    PDEBUG(DBG_SCRIPT, "result %04x\n", retval);
    offset += FIRMWARE_STEP_SIZE;
    pbuf += FIRMWARE_STEP_SIZE;
  }
  count = size + FIRMWARE_STEP_SIZE;
  PDEBUG(DBG_SCRIPT, "count %04x %d\n", offset, count);

  retval = usb_control_msg(cam->udev, usb_sndctrlpipe(udev, 0),
                                0x04, 0x40, 0x0015, offset, pbuf, count, 100);
  PDEBUG(DBG_SCRIPT, "result %d\n", retval);

  rbuf = buff;
  retval = usb_control_msg(cam->udev, usb_rcvctrlpipe(udev, 0),
                                0x04, 0xc0, 0x0016, offset, buff, 0x10, 100);
  PDEBUG(DBG_SCRIPT, "result %d\n", retval);
  for (i = 0; i < 5; i++) {
    PDEBUG(DBG_SCRIPT, "read result %02x %02x ", *(rbuf++), fwbuf[offset + i]);
  }

  release_firmware(firmware);
  kfree(buff);
  kfree(fwbuf);

  return (retval > 0 ? 0 : -1);
}

char *strdup(char *name) {
  char *copy;

  copy = kmalloc(strlen(name), GFP_KERNEL);
  if (copy != NULL)
    strcpy(copy, name);

  return copy;
}

struct nlist *install(char *name, unsigned short entry_value,
                      struct nlist **hashtab)
{
  struct nlist *np;
  unsigned hashval;

//   PDEBUG(DBG_SCRIPT, "install %s %04x", name, entry_value);
  if ((np = lookup(name, hashtab)) == NULL) {
    if (!(np = (struct nlist *) kmalloc(sizeof(*np), GFP_KERNEL)) )
      return NULL;

    if ( !(np->name = strdup(name)) ) {
      kfree (np);
      return NULL;
    }
   
    hashval = hash(name);
    np->next = hashtab[hashval];
    hashtab[hashval] = np;
  }
  np->entry_value = entry_value;
  return np;
}

void clear_hashtab(struct nlist **hashtab)
{
  int i;
  struct nlist *np;
  int empty = 0;

  for (i = 0; i < HASHSIZE; i++ ) {
    np = hashtab[i];
    if (np == NULL)
      empty++;
//     PDEBUG(DBG_SCRIPT, "hashindex %d", i);
    for (np = hashtab[i];np != NULL;np = np->next) {
      if (np->name != NULL) {
//         PDEBUG(DBG_SCRIPT, "clear entry %s", np->name);
        kfree((void *) np->name);
      }
      kfree(np);
    }
  }
  PDEBUG(DBG_SCRIPT, "empty nodes %d", empty);
}



static int init_hashtab(struct m560x_camera *cam) {
  const struct firmware *firmware;
  u8 *fw_end;
  u8 *line;
  int retval = 0;
  int n;
  u8 *entry_name;
  unsigned int  entry_value;
  int tokens[20];
  unsigned char type_id;
  struct nlist **hashtab = cam->hashtab;

  PDEBUG(DBG_SCRIPT, "Initialize hashtab with ali register names");
  retval = request_firmware(&firmware, M560X_INF_FILENAME, &cam->udev->dev);
  if(retval) {
    PDEBUG(DBG_SCRIPT, "request_firmware failed: " M560X_INF_FILENAME );
    return retval;
  }

  PDEBUG(DBG_SCRIPT, "Size of script file %zd", firmware->size);
  fw_end = firmware->data + firmware->size;

  line = script_find_section(firmware->data, firmware->size, 
                                                "SensorRegSetting");
  if(!line) {
    retval = -EINVAL;
    goto fail;
  }


  while ((line = script_next_line(line, fw_end - line))) {
    type_id = line[0];
    if(type_id == '[')
    	break;
    n = getTokens(tokens, line, 15);
    if (n >= 5) {
      entry_name = line + tokens[2];
      entry_value =  (unsigned int)simple_strtoul(line + tokens[4], NULL, 10);
      install(entry_name, entry_value, hashtab);
    }
  }

fail:
  release_firmware(firmware);
  return retval;
}

static int initialize_v4l_camera(struct m560x_camera *cam) {
  struct m560x_sensor *s = cam->sensor;
  struct v4l2_control ctrl;
//   struct v4l2_queryctrl *qctrl;
//   struct v4l2_rect *rect;
  u8 i = 0, n = 0;
  int err = 0;
//   if (!(cam->state & DEV_INITIALIZED)) {
//     init_waitqueue_head(&cam->open);
//     qctrl = s->qctrl;
//     rect = &(s->cropcap.defrect);
//   } else { /* use current values */
//     qctrl = s->_qctrl;
//     rect = &(s->_rect);
//   }
  s->init(cam);

  n = sizeof(s->qctrl) / sizeof(s->qctrl[0]);
  PDEBUG(DBG_TRACE, "initialize controls to default values %zd %zd %d %s",
                              sizeof(s->qctrl), sizeof(s->qctrl[0]),
                              n, s->name);
  for (i = 0; i < n; i++) {
    PDEBUG(DBG_TRACE, "i %d", i);
    if (s->qctrl[i].id != 0 &&
              !(s->qctrl[i].flags & V4L2_CTRL_FLAG_DISABLED)) {
      ctrl.id = s->qctrl[i].id;
      ctrl.value = s->qctrl[i].default_value;
      err = s->set_ctrl(cam, &ctrl);
      if (err) {
        PDEBUG(DBG_TRACE, "Set %s control failed", s->qctrl[i].name);
        return err;
      }
     PDEBUG(DBG_TRACE, "Image sensor supports '%s' control",
     s->qctrl[i].name);
    }
  }
  return 0;
}

static int initialize_m5602_camera(struct m560x_camera *cam) {
  int result = 0;
  int i;

  PDEBUG(DBG_TRACE, "initialize_m5602_camera");

  if (!(cam->state & DEV_INITIALIZED)) {
    init_waitqueue_head(&cam->open);
    init_MUTEX(&cam->fileop_sem);
    spin_lock_init(&cam->queue_lock);
    init_waitqueue_head(&cam->wait_frame);
    init_waitqueue_head(&cam->wait_stream);
    cam->nreadbuffers = 2;
    cam->state |= DEV_INITIALIZED;
  }

  result = read_m5602_register(cam, 0x8140);
  PDEBUG(DBG_SCRIPT, "result %d 0x%02x\n", result, result);
  result = read_m5602_register(cam, 0x8140);
  PDEBUG(DBG_SCRIPT, "result %d 0x%02x\n", result, result);

//  urb 312 = 604
  m5602_init_from_script(cam, "INITCAM");
  m5602_init_from_script(cam, "STOPCAM");
  m5602_init_from_script(cam, "STARTCAM");
  m5602_init_from_script(cam, "INIT4");
  m5602_init_from_script(cam, "INIT5");
  m5602_init_from_script(cam, "STARTSENSOR");
  m5602_init_from_script(cam, "SETSIZE1");
  m5602_init_from_script(cam, "SETSIZE2");
  m5602_init_from_script(cam, "SENSORCLOCKA");
  m5602_init_from_script(cam, "SETBRIGHT");
  m5602_init_from_script(cam, "SETSIZE2");
  m5602_init_from_script(cam, "SENSORCLOCKB");
  m5602_init_from_script(cam, "INIT6");

  // urb 1288 1475
  m5602_init_from_script(cam, "STOPCAM");
  usb_set_interface(cam->udev, 0, 1);
  result = read_m5602_register(cam, 0xffb);
  PDEBUG(DBG_SCRIPT, "result %d 0x%02x\n", result, result);
  m5602_init_from_script(cam, "STARTCAM");
  m5602_init_from_script(cam, "INIT4");
  m5602_init_from_script(cam, "INIT5");
  m5602_init_from_script(cam, "STARTSENSOR");
  m5602_init_from_script(cam, "SETSIZE1");
  m5602_init_from_script(cam, "SETSIZE2");
  m5602_init_from_script(cam, "SENSORCLOCKF");
  m5602_init_from_script(cam, "SETBRIGHT");
  m5602_init_from_script(cam, "SENSORCLOCKF");
  write_m5602_sensor_register(cam, 0xba, 0x09, 0x01, 0xa1);
  write_m5602_sensor_register(cam, 0xba, 0x2f, 0x07, 0xff);
  write_m5602_register(cam, 0x8176, 0x5b);
  write_m5602_sensor_register(cam, 0xba, 0xf0, 0x00, 0x00);
  write_m5602_sensor_register(cam, 0xba, 0x20, 0x01, 0x03);
  for (i = 0; i < 20; i++) {
    result = write_m5602_register(cam, 0xFF9, 0x01);
    PDEBUG(DBG_TRACE, "0x0ff9 %02x", result);
  }

  return 0;
}


static int initialize_m5603_camera(struct m560x_camera *cam)
{
  struct usb_device *udev = cam->udev;
  u8 *buff = cam->control_buffer;
  int result = 0;
  int initialize;

  PDEBUG(DBG_TRACE, "initialize_m5603_camera");

  if (!(cam->state & DEV_INITIALIZED)) {
    init_waitqueue_head(&cam->open);
    init_MUTEX(&cam->fileop_sem);
    spin_lock_init(&cam->queue_lock);
    init_waitqueue_head(&cam->wait_frame);
    init_waitqueue_head(&cam->wait_stream);
    cam->nreadbuffers = 2;
    cam->state |= DEV_INITIALIZED;
  }


//   halt_dsp(cam);
  result = or_ali_register(cam, 0x8000, 0x00);
  PDEBUG(DBG_SCRIPT, "0x8000 %02x", result);
  result = read_ali_register(cam, 0x8140);
  PDEBUG(DBG_SCRIPT, "0x8140 %02x", result);
  result = read_ali_register(cam, 0x8140);
  PDEBUG(DBG_SCRIPT, "0x8140 %02x", result);
  initialize = result;
  result = read_ali_register(cam, 0x8170);
  PDEBUG(DBG_SCRIPT, "result %d 0x%02x\n", result, result);
  result = and_ali_register(cam, 0x8173, 0x00);     // xbGPIODATAL
  PDEBUG(DBG_SCRIPT, "result %d 0x%02x\n", result, result);
  result = read_ali_register(cam, 0x8175);         // xbGPIODATAL
  PDEBUG(DBG_SCRIPT, "result %d 0x%02x\n", result, result);
  result = write_ali_register(cam, 0x8173, 0x11);
  PDEBUG(DBG_SCRIPT, "result %d 0x%02x\n", result, result);
  result = and_ali_register(cam, 0x8172, 0x00);     // xbGPIODIRH
  PDEBUG(DBG_SCRIPT, "result %d 0x%02x\n", result, result);
  result = read_ali_register(cam, 0x8174);         // xbGPIODATAH
  PDEBUG(DBG_SCRIPT, "result %d 0x%02x\n", result, result);
  result = write_ali_register(cam, 0x8172, 0x21);
  PDEBUG(DBG_SCRIPT, "result %d 0x%02x\n", result, result);
  result = and_ali_register(cam, 0x8140, 0xfb);
  PDEBUG(DBG_SCRIPT, "result %d 0x%02x\n", result, result);
  if (initialize == 0x86) {
    result = load_firmware(cam);
    if(result)
      return result;
    result = usb_control_msg(udev, usb_sndctrlpipe(udev, 0),
                                0x04, 0x40, 0x10, 0x0000, buff, 1, 100);
    PDEBUG(DBG_SCRIPT, "register %04x result %02x value %02x",
                                0x0000, result, *buff);
  }
  init_hashtab(cam);
//   restart_dsp(cam);

  result = m5603_init_from_script(cam, "DSP_INIT");
  if (result < 0)  {
    err("error from m5603_init_from_script");
    return result;
  }
  m5603_init_from_script(cam, "AP_ON");
  m5603_init_from_script(cam, "R_VGA");
//   m5603_init_from_script(cam, "EXTRA3");

  return 0;
}

void m560x_attach_sensor(struct m560x_camera *cam,
                         struct m560x_sensor *sensor)
{
  PDEBUG(DBG_TRACE, "attach sensor");
  cam->sensor = sensor;
  cam->sensor->dev = &cam->dev;
  cam->sensor->usbdev = cam->udev;
}

static int initialize_camera(struct m560x_camera *cam) {
  int result = 0;

  switch (cam->camera_type) {
  case CAMERA_M5602:
    result = initialize_m5602_camera(cam);
    break;

  case CAMERA_M5603:
    result = initialize_m5603_camera(cam);
    break;

  default:
    result = -1;
    break;
  }

  if (!result)
    PDEBUG(DBG_TRACE, "Correctly initialized_m560x_camera");

  return result;
}


/******************* isochronous data management ******************************/

/* Bytes per frame. Used by read(). Return of 0 indicates error */
static inline long int
    get_frame_length(struct m560x_frame_t *frame)
{
  if (!frame)
    return 0;
  else
    return (3 *frame->width * frame->height);
}

typedef enum
{
  BAYER_PATTERN_BGGR,
  BAYER_PATTERN_GRBG,
  BAYER_PATTERN_RGGB,
  BAYER_PATTERN_GBRG
} bayer_pattern_t;

void BayerNearestNeighbor( unsigned char *src,
                           unsigned char *dest,
                           int sx,
                           int sy,
                           bayer_pattern_t type)
{
  unsigned char *outR, *outG, *outB;
  register int i, j;
 // sx and sy should be even
  switch (type) {
  case BAYER_PATTERN_GRBG:
  case BAYER_PATTERN_BGGR:
    outR=&dest[0];
    outG=&dest[1];
    outB=&dest[2];
    break;
  case BAYER_PATTERN_GBRG:
  case BAYER_PATTERN_RGGB:
    outR=&dest[2];
    outG=&dest[1];
    outB=&dest[0];
    break;
  default:
    outR=NULL;outG=NULL;outB=NULL;
    break;
  }

  switch (type) {
  case BAYER_PATTERN_GRBG: //-------------------------------------------
  case BAYER_PATTERN_GBRG:
 // copy original RGB data to output images
    for (i=0;i<sy;i+=2) {
      for (j=0;j<sx;j+=2) {
        outG[(i*sx+j)*3]=src[i*sx+j];
        outG[((i+1)*sx+(j+1))*3]=src[(i+1)*sx+(j+1)];
        outR[(i*sx+j+1)*3]=src[i*sx+j+1];
        outB[((i+1)*sx+j)*3]=src[(i+1)*sx+j];
    //    printf("%d %d %d %d %08x %08x\n", i, j, ((i+1)*sx+j)*3, (i+1)*sx+(j+1), ((i+1)*sx+j)*3, (i+1)*sx+(j+1));
      }
    }
   // R channel
    for (i=0;i<sy;i+=2) {
      for (j=0;j<sx-1;j+=2) {
        outR[(i*sx+j)*3]=outR[(i*sx+j+1)*3];
        outR[((i+1)*sx+j+1)*3]=outR[(i*sx+j+1)*3];
        outR[((i+1)*sx+j)*3]=outR[(i*sx+j+1)*3];
      }
    }
    // B channel
    for (i=0;i<sy-1;i+=2)  { //every two lines
      for (j=0;j<sx-1;j+=2) {
        outB[(i*sx+j)*3]=outB[((i+1)*sx+j)*3];
        outB[(i*sx+j+1)*3]=outB[((i+1)*sx+j)*3];
        outB[((i+1)*sx+j+1)*3]=outB[((i+1)*sx+j)*3];
      }
    }
  // using lower direction for G channel

  // G channel
    for (i=0;i<sy-1;i+=2)//every two lines
      for (j=1;j<sx;j+=2)
        outG[(i*sx+j)*3]=outG[((i+1)*sx+j)*3];

    for (i=1;i<sy-2;i+=2)//every two lines
      for (j=0;j<sx-1;j+=2)
        outG[(i*sx+j)*3]=outG[((i+1)*sx+j)*3];

  // copy it for the next line
    for (j=0;j<sx-1;j+=2)
      outG[((sy-1)*sx+j)*3]=outG[((sy-2)*sx+j)*3];

    break;
    case BAYER_PATTERN_BGGR: //-------------------------------------------
  case BAYER_PATTERN_RGGB:
  // copy original data
    for (i=0;i<sy;i+=2) {
      for (j=0;j<sx;j+=2) {
        outB[(i*sx+j)*3]=src[i*sx+j];
        outR[((i+1)*sx+(j+1))*3]=src[(i+1)*sx+(j+1)];
        outG[(i*sx+j+1)*3]=src[i*sx+j+1];
        outG[((i+1)*sx+j)*3]=src[(i+1)*sx+j];
      }
    }
  // R channel
    for (i=0;i<sy;i+=2){
      for (j=0;j<sx-1;j+=2) {
        outR[(i*sx+j)*3]=outR[((i+1)*sx+j+1)*3];
        outR[(i*sx+j+1)*3]=outR[((i+1)*sx+j+1)*3];
        outR[((i+1)*sx+j)*3]=outR[((i+1)*sx+j+1)*3];
      }
    }
  // B channel
    for (i=0;i<sy-1;i+=2) { //every two lines
      for (j=0;j<sx-1;j+=2) {
        outB[((i+1)*sx+j)*3]=outB[(i*sx+j)*3];
        outB[(i*sx+j+1)*3]=outB[(i*sx+j)*3];
        outB[((i+1)*sx+j+1)*3]=outB[(i*sx+j)*3];
      }
    }
  // using lower direction for G channel

  // G channel
    for (i=0;i<sy-1;i+=2)//every two lines
      for (j=0;j<sx-1;j+=2)
        outG[(i*sx+j)*3]=outG[((i+1)*sx+j)*3];

    for (i=1;i<sy-2;i+=2)//every two lines
      for (j=0;j<sx-1;j+=2)
        outG[(i*sx+j+1)*3]=outG[((i+1)*sx+j+1)*3];

  // copy it for the next line
    for (j=0;j<sx-1;j+=2)
      outG[((sy-1)*sx+j+1)*3]=outG[((sy-2)*sx+j+1)*3];

    break;

    default:  //-------------------------------------------
      break;
  }
}



static void *m560x_find_sof_header(struct m560x_camera *cam, void *mem,
                                   size_t len)
{
  size_t soflen = 2;
  size_t i;
  u8 header[2] = {0xff, 0xff};

  for (i = 0;(len >= soflen) && (i <= len - soflen) ; i++) {
    if (!memcmp(mem + i, header, 2)) {
      return mem + i + soflen;
    }
  }
  return NULL;
}


static void *m560x_find_eof_header(struct m560x_camera *cam, void *mem,
                                   size_t len)
{
  return NULL;
}

static void isoc_urb_complete(struct urb * urb) {
  struct m560x_camera * cam = urb->context;
  struct m560x_frame_t ** f;
  size_t buffersize;

  unsigned long lock_flags;
  u8 i;
  int err = 0;



  if (cam->stream != STREAM_ON)
    PDEBUG(DBG_DATA, "urb completion routine %d", cam->stream);

  if (urb->status == -ENOENT)
    return;

  f = &cam->frame_current;


  if (cam->stream == STREAM_INTERRUPT) {
    cam->stream = STREAM_OFF;
    if ((*f))
      (*f)->state = FRAME_QUEUED;
    PDEBUG(DBG_TRACE, "Stream interrupted");
        wake_up_interruptible(&cam->wait_stream);
  }

  if (cam->state & DEV_DISCONNECTED)
    return;

  if (cam->state & DEV_MISCONFIGURED) {
    wake_up_interruptible(&cam->wait_frame);
    return;
  }

  if (cam->stream == STREAM_OFF || list_empty(&cam->inqueue))
    goto resubmit_urb;

  if (!(*f))
    (*f) = list_entry(cam->inqueue.next, struct m560x_frame_t,
  frame);


  buffersize = 0x50000;

  for (i = 0; i < urb->number_of_packets; i++) {
    unsigned int img, len, status;
    void *pos, *sof, *eof;

    len = urb->iso_frame_desc[i].actual_length;
    status = urb->iso_frame_desc[i].status;
    pos = urb->iso_frame_desc[i].offset + urb->transfer_buffer;
    if (status != 0)
      PDEBUG(DBG_TRACE, "urb packet %d status %d", i, status);

    PDEBUG(DBG_DATA, "packet length %08x status %d pos %p", len, status, pos);
    PDEBUG(DBG_DATA, "header %02x %02x %02x %02x %02x %02x %02x %02x",
                                ((unsigned char *)pos)[0],
                                ((unsigned char *)pos)[1],
                                ((unsigned char *)pos)[2],
                                ((unsigned char *)pos)[3],
                                ((unsigned char *)pos)[4],
                                ((unsigned char *)pos)[5],
                                ((unsigned char *)pos)[6],
                                ((unsigned char *)pos)[7]);
    if (status) {
      PDEBUG(DBG_TRACE, "Error in isochronous frame");
          (*f)->state = FRAME_ERROR;
      continue;
    }

    if (len <= 4)
      continue;

    /* strip usb packet header ff ff ff ff */
    len = len - 4;
    pos = pos + 4;



    /*
    NOTE: It is probably correct to assume that SOF and EOF
    headers do not occur between two consecutive packets,
    but who knows..Whatever is the truth, this assumption
    doesn't introduce bugs.
    */

redo:
    sof = m560x_find_sof_header(cam, pos , len);
    if (!sof ) {
      eof = m560x_find_eof_header(cam, pos, len);
      if ((*f)->state == FRAME_GRABBING) {

        img = len;

        if (eof)
          img = (eof > pos) ? eof - pos - 1 : 0;


        if ((*f)->buf.bytesused+img > buffersize) {
          u32 b = (*f)->buf.bytesused + img - buffersize;
          img = buffersize - (*f)->buf.bytesused;
          PDEBUG(DBG_TRACE, "Expected EOF not found: video frame cut");
          if (eof)
            PDEBUG(DBG_TRACE, "Exceeded limit: +%u "
              "bytes", (unsigned)(b));
        }


        memcpy((*f)->rawdata + (*f)->buf.bytesused, pos,
                 img);

        if ((*f)->buf.bytesused == 0)
          do_gettimeofday(&(*f)->buf.timestamp);

        (*f)->buf.bytesused += img;

end_of_frame:
        if (((*f)->buf.bytesused == buffersize) || (sof != NULL)) {

          u32 b = (*f)->buf.bytesused;
          (*f)->state = FRAME_DONE;
          BayerNearestNeighbor( (*f)->rawdata, (*f)->bufmem, 640, 480,
                                  (bayer_pattern_t) BAYER_PATTERN_BGGR);
          (*f)->buf.bytesused = 3 * (*f)->buf.bytesused;
          (*f)->buf.sequence= ++cam->frame_count;
          spin_lock_irqsave(&cam->queue_lock, lock_flags);
          list_move_tail(&(*f)->frame, &cam->outqueue);
          if (!list_empty(&cam->inqueue))
            (*f) = list_entry(cam->inqueue.next, struct m560x_frame_t, frame);
          else
            (*f) = NULL;
          spin_unlock_irqrestore(&cam->queue_lock, lock_flags);

          PDEBUG(DBG_TRACE, "Video frame captured: %lu bytes",
                                                        (unsigned long)(b));

          if (!(*f))
            goto resubmit_urb;

        } else if (eof) {
          (*f)->state = FRAME_ERROR;
          PDEBUG(DBG_TRACE, "Not expected EOF after %lu "
                    "bytes of image data",
                    (unsigned long)((*f)->buf.bytesused));
        }

        if (sof) /* (1) */
          goto start_of_frame;

      } else if (eof) {
        PDEBUG(DBG_TRACE, "EOF without SOF");
        continue;

      } else {
//         PDEBUG(DBG_TRACE, "Ignoring pointless isochronous frame");
        continue;
      }
    } else if ((*f)->state == FRAME_QUEUED || (*f)->state == FRAME_ERROR) {
start_of_frame:
      (*f)->state = FRAME_GRABBING;
      (*f)->buf.bytesused = 0;
      len -= (sof - pos);
      pos = sof;
      PDEBUG(DBG_TRACE, "SOF detected: new video frame");
      if (len)
        goto redo;
    } else if ((*f)->state == FRAME_GRABBING) {


      PDEBUG(DBG_TRACE, "SOF detected: FRAME_GRABBING %02tx", (sof - pos));
      goto end_of_frame; /* (1) */

    }
  }


resubmit_urb:
  urb->dev = cam->udev;
  err = usb_submit_urb(urb, GFP_ATOMIC);
  if (err < 0 && err != -EPERM) {
    cam->state |= DEV_MISCONFIGURED;
    PDEBUG(DBG_TRACE, "usb_submit_urb() failed");
  }

  wake_up_interruptible(&cam->wait_frame);
}



static int m560x_start_transfer(struct m560x_camera *cam)
{
  int retval = 0;
  int i, j;
  int result = 0;
  struct urb * urb;
  struct usb_device *udev = cam->udev;


  PDEBUG(DBG_TRACE, "start_transfer");
  PDEBUG(DBG_TRACE, "Initialize isochronous buffers and urb's");
  PDEBUG(DBG_TRACE, "number of urb's %d packetsize %04x number of packets %d",
                                NB_ISOC_URBS, ISOC_PKTSIZE, NB_ISOC_PACKETS);

  cam->isoc_in_size = ISOC_PKTSIZE;
  for (i = 0; i < NB_ISOC_URBS; i++) {
    cam->isoc_in_buffer[i] = kmalloc(NB_ISOC_PACKETS*ISOC_PKTSIZE, GFP_KERNEL);
    if (!cam->isoc_in_buffer[i]) {
      retval = -ENOMEM;
      err("no memory for isoc buffer");
    }
  }

  for (i = 0; i < NB_ISOC_URBS; i++) {
    urb = usb_alloc_urb(NB_ISOC_PACKETS, GFP_KERNEL);
    if (!urb) {
      retval = -ENOMEM;
      err("no memory for isoc buffer");
    }
    cam->urb[i] = urb;
    urb->dev = udev;
    urb->context = cam;
    urb->pipe = usb_rcvisocpipe(udev, 1);
    urb->transfer_flags = URB_ISO_ASAP;
    urb->number_of_packets = NB_ISOC_PACKETS;
    urb->complete = isoc_urb_complete;
    urb->transfer_buffer = cam->isoc_in_buffer[i];
    urb->transfer_buffer_length = ISOC_PKTSIZE * NB_ISOC_PACKETS;
    urb->interval = 1;
    for (j = 0; j < NB_ISOC_PACKETS; j++) {
      urb->iso_frame_desc[j].offset = ISOC_PKTSIZE * j;
      urb->iso_frame_desc[j].length = ISOC_PKTSIZE;
      urb->iso_frame_desc[j].actual_length = 0;
      urb->iso_frame_desc[j].status = 0;
    }
  }
  PDEBUG(DBG_TRACE, "Isochronous buffers and urb's initialized");
  PDEBUG(DBG_TRACE, "Bayer_frame pointer %p", cam->bayer_frame);

  if (cam->camera_type == CAMERA_M5603) {
    usb_set_interface(cam->udev, 0, 1);
    for (i = 0; i < 10; i++) {
      result = write_ali_register(cam, 0xFF9, 0x01);
      PDEBUG(DBG_TRACE, "0x0ff9 %02x", result);
    }
  }

  cam->frame_current = NULL;

  for (i = 0; i < NB_ISOC_URBS; i++) {
    if (cam->urb[i]) {
      retval = usb_submit_urb(cam->urb[i], GFP_KERNEL);
      if (retval) {
        err("submit of urb failed %d", retval);
      }
    }
  }

  return retval;
}


static int m560x_stop_transfer(struct m560x_camera *cam)
{
  int retval = 0;
  int i;


  PDEBUG(DBG_TRACE, "Free isochronous buffers and urb's");

  for (i = NB_ISOC_URBS - 1; i >= 0; i--) {
    usb_kill_urb(cam->urb[i]);
    usb_free_urb(cam->urb[i]);
    kfree(cam->isoc_in_buffer[i]);
  }

  PDEBUG(DBG_TRACE, "Isochronous buffers and urb's freed");
  usb_set_interface(cam->udev, 0, 0);



  return retval;
}

static int m560x_stream_interrupt(struct m560x_camera *cam)
{
  int err = 0;

  cam->stream = STREAM_INTERRUPT;
  err = wait_event_timeout(cam->wait_stream,
                           (cam->stream == STREAM_OFF) ||
                               (cam->state & DEV_DISCONNECTED),
                           M560X_URB_TIMEOUT);
  if (cam->state & DEV_DISCONNECTED)
    return -ENODEV;
  else if (err) {
    cam->state |= DEV_MISCONFIGURED;
    PDEBUG(DBG_TRACE, "The camera is misconfigured. To use it, close and "
        "open /dev/video%d again.", cam->vdev->minor);
    return err;
  }

  return 0;
}

/****************************************************************************
 *  sysfs
 ***************************************************************************/
#define m560x_file(name, variable, field)                              \
static ssize_t show_##name(struct class_device *class_dev, char *buf)   \
{                                                                       \
        struct video_device *vdev = to_video_device(class_dev);         \
        struct m560x_camera *cam = video_get_drvdata(vdev);                  \
        return sprintf(buf, field, cam->variable);                      \
}                                                                       \
static CLASS_DEVICE_ATTR(name, S_IRUGO, show_##name, NULL);

m560x_file(model, name, "%s\n");
m560x_file(in_use, user, "%d\n");
m560x_file(streaming, streaming, "%d\n");
m560x_file(palette, palette, "%i\n");
m560x_file(frames_total, readCount, "%d\n");
m560x_file(frames_read, frameCount, "%d\n");
m560x_file(packets_dropped, dropped, "%d\n");
m560x_file(decoding_errors, error, "%d\n");

static void m560x_create_sysfs_files(struct video_device *vdev)
{
  video_device_create_file(vdev, &class_device_attr_model);
  video_device_create_file(vdev, &class_device_attr_in_use);
  video_device_create_file(vdev, &class_device_attr_streaming);
  video_device_create_file(vdev, &class_device_attr_palette);
  video_device_create_file(vdev, &class_device_attr_frames_total);
  video_device_create_file(vdev, &class_device_attr_frames_read);
  video_device_create_file(vdev, &class_device_attr_packets_dropped);
  video_device_create_file(vdev, &class_device_attr_decoding_errors);
}

static void m560x_remove_sysfs_files(struct video_device *vdev)
{
  video_device_remove_file(vdev, &class_device_attr_model);
  video_device_remove_file(vdev, &class_device_attr_in_use);
  video_device_remove_file(vdev, &class_device_attr_streaming);
  video_device_remove_file(vdev, &class_device_attr_palette);
  video_device_remove_file(vdev, &class_device_attr_frames_total);
  video_device_remove_file(vdev, &class_device_attr_frames_read);
  video_device_remove_file(vdev, &class_device_attr_packets_dropped);
  video_device_remove_file(vdev, &class_device_attr_decoding_errors);
}
/*
 *  ---------------------------------------------------------------------------
 * Video4Linux Functions
 * ---------------------------------------------------------------------------
 */
static void m560x_vm_open(struct vm_area_struct *vma)
{
  struct m560x_frame_t *f = vma->vm_private_data;
  f->vma_use_count++;
}


static void m560x_vm_close(struct vm_area_struct *vma)
{
  /* NOTE: buffers are not freed here */
  struct m560x_frame_t *f = vma->vm_private_data;
  f->vma_use_count--;
}


static struct vm_operations_struct m560x_vm_ops = {
  .open = m560x_vm_open,
  .close = m560x_vm_close,
};

static void m560x_release_resources(struct m560x_camera *cam)
{
  down(&m560x_sysfs_lock);

  info("m560x release resources video device %d", cam->vdev->minor);
  video_set_drvdata(cam->vdev, NULL);
  video_unregister_device(cam->vdev);
  m560x_remove_sysfs_files(cam->vdev);

  up(&m560x_sysfs_lock);

  if (cam->control_buffer)
    kfree(cam->control_buffer);
  if (cam->hashtab) {
    clear_hashtab(cam->hashtab);
    kfree(cam->hashtab);
  }
}

static int v4l_m560x_open(struct inode *inode, struct file *filp)
{
  struct m560x_camera *cam;
  int err = 0;

  /*
  This is the only safe way to prevent race conditions with
  disconnect
  */

  if (!down_read_trylock(&m560x_disconnect))
    return -ERESTARTSYS;

  cam = video_get_drvdata(video_devdata(filp));
  PDEBUG(DBG_TRACE, "Device /dev/video%d ", cam->vdev->minor);
  info("m560x open");

  if (down_interruptible(&cam->dev_sem)) {
    up_read(&m560x_disconnect);
    return -ERESTARTSYS;
  }
  PDEBUG(DBG_TRACE, "cam->users %d ", cam->users);

  if (cam->users) {
    PDEBUG(DBG_TRACE, "Device /dev/video%d is busy...", cam->vdev->minor);
    if ((filp->f_flags & O_NONBLOCK) ||
             (filp->f_flags & O_NDELAY)) {
      err = -EWOULDBLOCK;
      goto out;
    }
    up(&cam->dev_sem);
    err = wait_event_interruptible_exclusive(cam->open,
                                                cam->state & DEV_DISCONNECTED
                                                || !cam->users);
    if (err) {
      up_read(&m560x_disconnect);
      return err;
    }
    if (cam->state & DEV_DISCONNECTED) {
      up_read(&m560x_disconnect);
      return -ENODEV;
    }
    down(&cam->dev_sem);
  }


  if (cam->state & DEV_MISCONFIGURED) {
    err = initialize_camera(cam);
    if (err) {
      PDEBUG(DBG_TRACE, "Initialization failed again. "
                        "I will retry on next open().");
      goto out;
    }
    cam->state &= ~DEV_MISCONFIGURED;
  }


  if ((err = m560x_start_transfer(cam)))
    goto out;

  filp->private_data = cam;
  cam->users++;
  cam->io = IO_NONE;
  cam->stream = STREAM_OFF;
  cam->nbuffers = 0;
  cam->frame_count = 0;
  m560x_empty_framequeues(cam);

  PDEBUG(DBG_TRACE, "cam %p", cam);
  PDEBUG(DBG_TRACE, "Video device /dev/video%d is open", cam->vdev->minor);

out:
  up(&cam->dev_sem);
  up_read(&m560x_disconnect);

  return err;

}

static int v4l_m560x_release(struct inode *inode, struct file *filp)
{
  struct m560x_camera *cam = video_get_drvdata(video_devdata(filp));
  PDEBUG(DBG_TRACE, "m560x release");

  down(&cam->dev_sem); /* prevent disconnect() to be called */

  m560x_stop_transfer(cam);
  PDEBUG(DBG_TRACE, "cam %p", cam);
  PDEBUG(DBG_TRACE, "nbuffers %d", cam->nbuffers);
  m560x_release_buffers(cam);

  cam->users--;
//   wake_up_interruptible_nr(&cam->open, 1);

  PDEBUG(DBG_TRACE, "Video device /dev/video%d closed", cam->vdev->minor);

  up(&cam->dev_sem);

  return 0;

}

static ssize_t v4l_m560x_read(struct file *filp, char __user *buf,
                              size_t count, loff_t *f_pos)
{
  struct m560x_camera * cam = video_get_drvdata(video_devdata(filp));
  struct m560x_frame_t *f, * i;
  unsigned long lock_flags;
  int err = 0;

  PDEBUG(DBG_TRACE, "Read call");

  if (down_interruptible(&cam->fileop_sem))
    return -ERESTARTSYS;

  if (cam->state & DEV_DISCONNECTED) {
    PDEBUG(DBG_TRACE, "Device not present");
    err = -ENODEV;
    goto out;
  }

  if (cam->state & DEV_MISCONFIGURED) {
    PDEBUG(DBG_TRACE, "The camera is misconfigured. Close and open it again.");
    err = -EIO;
    goto out;
  }

  if (cam->io == IO_MMAP) {
    PDEBUG(DBG_TRACE, "Close and open the device again to choose "
        "the read method");
    err = -EINVAL;
    goto out;
  }

  PDEBUG(DBG_TRACE, "io_method %d", cam->io);

  if (cam->io == IO_NONE) {
    PDEBUG(DBG_TRACE, "IO_NONE");
    if (!m560x_request_buffers(cam, cam->nreadbuffers, IO_READ)) {
      PDEBUG(DBG_TRACE, "read() failed, not enough memory");
      err = -ENOMEM;
      goto out;
    }
    cam->io = IO_READ;
    cam->stream = STREAM_ON;
    PDEBUG(DBG_TRACE, "set IO_READ");
    m560x_queue_unusedframes(cam);
  }

  if (!count) {
    err = 0;
    goto out;
  }

  if (list_empty(&cam->outqueue)) {
    if (filp->f_flags & O_NONBLOCK) {
      err = -EAGAIN;
      goto out;
    }
    err = wait_event_interruptible(cam->wait_frame,
                        ((!list_empty(&cam->outqueue)) ||
                        (cam->state & (DEV_DISCONNECTED|DEV_MISCONFIGURED))) );
    if (err)
      goto out;
    if (cam->state & DEV_DISCONNECTED) {
      err = -ENODEV;
      goto out;
    }
    if (cam->state & DEV_MISCONFIGURED) {
      err = -EIO;
      goto out;
    }
  }

  f = list_entry(cam->outqueue.prev, struct m560x_frame_t, frame);

  spin_lock_irqsave(&cam->queue_lock, lock_flags);
  list_for_each_entry(i, &cam->outqueue, frame)
  i->state = FRAME_UNUSED;
  INIT_LIST_HEAD(&cam->outqueue);
  spin_unlock_irqrestore(&cam->queue_lock, lock_flags);

  m560x_queue_unusedframes(cam);

  if (count > f->buf.bytesused)
    count = f->buf.bytesused;

  if (copy_to_user(buf, f->bufmem, count)) {
    err = -EFAULT;
    goto out;
  }
  *f_pos += count;

  PDEBUG(DBG_TRACE, "Frame #%lu, bytes read: %zu",
                                (unsigned long)f->buf.index, count);

out:
  up(&cam->fileop_sem);
  return (err) ? err : count;
}

static unsigned int v4l_m560x_poll(struct file *filp, poll_table *wait)
{
  struct m560x_camera *cam = video_get_drvdata(video_devdata(filp));
  unsigned int mask = 0;

//   PDEBUG(DBG_TRACE, "v4l_m560x_poll");
  if (down_interruptible(&cam->fileop_sem))
    return POLLERR;

  if (cam->state & DEV_DISCONNECTED) {
    PDEBUG(DBG_TRACE, "Device not present");
    goto error;
  }

  if (cam->state & DEV_MISCONFIGURED) {
    PDEBUG(DBG_TRACE, "The camera is misconfigured. Close and open it again.");
    goto error;
  }

  if (cam->io == IO_NONE) {
    PDEBUG(DBG_TRACE, "IO_NONE");
    if (!m560x_request_buffers(cam, cam->nreadbuffers,
         IO_READ)) {
           PDEBUG(DBG_TRACE, "poll() failed, not enough memory");
           goto error;
         }
         cam->io = IO_READ;
         cam->stream = STREAM_ON;
  }

  if (cam->io == IO_READ) {
    PDEBUG(DBG_TRACE, "IO_READ");
    m560x_queue_unusedframes(cam);
  }

  poll_wait(filp, &cam->wait_frame, wait);

  if (!list_empty(&cam->outqueue)) {
    PDEBUG(DBG_TRACE, "!list_empty");
    mask |= POLLIN | POLLRDNORM;
  }

  up(&cam->fileop_sem);
  if (mask != 0)
    PDEBUG(DBG_TRACE, "v4l_m560x_poll return %08x", mask);

  return mask;

error:
  up(&cam->fileop_sem);
  PDEBUG(DBG_TRACE, "v4l_m560x_poll return error");
  return POLLERR;
}



static int v4l_m560x_mmap(struct file *filp, struct vm_area_struct *vma)
{
  struct m560x_camera *cam = video_get_drvdata(video_devdata(filp));
  unsigned long size = vma->vm_end - vma->vm_start;
  unsigned long start = vma->vm_start;
  unsigned long pos;
  unsigned long page;
  u32 i;
  int err = 0;

  PDEBUG(DBG_TRACE, "mmap call");
  if (down_interruptible(&cam->fileop_sem))
    return -ERESTARTSYS;

  if (cam->state & DEV_DISCONNECTED) {
    PDEBUG(DBG_TRACE, "Device not present");
    err = -ENODEV;
    goto out;
  }

  if (cam->state & DEV_MISCONFIGURED) {
    PDEBUG(DBG_TRACE, "The camera is misconfigured. Close and open it again.");
    err = -EIO;
    goto out;
  }

  if (cam->io != IO_MMAP || !(vma->vm_flags & VM_WRITE) ||
      size != PAGE_ALIGN(cam->frame[0].buf.length)) {
    err = -EINVAL;
    goto out;
  }

  for (i = 0; i < cam->nbuffers; i++) {
    if ((cam->frame[i].buf.m.offset>>PAGE_SHIFT) == vma->vm_pgoff)
      break;
  }
  if (i == cam->nbuffers) {
    err = -EINVAL;
    goto out;
  }

  /* VM_IO is eventually going to replace PageReserved altogether */
  vma->vm_flags |= VM_IO;
  vma->vm_flags |= VM_RESERVED; /* avoid to swap out this VMA */

  pos = (unsigned long)cam->frame[i].bufmem;
  while (size > 0) { /* size is page-aligned */
    page = vmalloc_to_pfn((void *)pos);
    if (remap_pfn_range(vma, start, page, PAGE_SIZE, vma->vm_page_prot)) {
      err = -EAGAIN;
      goto out;
    }
    start += PAGE_SIZE;
    pos += PAGE_SIZE;
    size -= PAGE_SIZE;
  }

  vma->vm_ops = &m560x_vm_ops;
  vma->vm_private_data = &cam->frame[i];

  m560x_vm_open(vma);

out:
  up(&cam->fileop_sem);

  return err;
}

static int v4l_m560x_do_ioctl(struct inode *inode, struct file *file,
                              unsigned int ioctlnr, void *arg)
{
  int retval = 0;
  struct m560x_camera *cam = (struct m560x_camera *)file->private_data;

  PDEBUG(DBG_TRACE, "v4l_m560x_ioctl %02x", (unsigned char) ioctlnr);

  if (!cam) {
    err("m560x video_device improperly initialized");
    retval = -ENODEV;
    goto exit;
  }

  switch (ioctlnr) {

  case VIDIOC_QUERYCAP:
  {
    struct v4l2_capability cap;

    PDEBUG(DBG_TRACE, "VIDIOC_QUERYCAP");
    memset(&cap, 0, sizeof(cap));
    strlcpy(cap.driver, "m560x", sizeof(cap.driver));
    cap.capabilities = V4L2_CAP_VIDEO_CAPTURE | V4L2_CAP_READWRITE
                                                        | V4L2_CAP_STREAMING;
//    cap->capabilities = V4L2_CAP_VIDEO_CAPTURE | V4L2_CAP_READWRITE ;
    cap.version = (__u32) DRIVER_VERSION_NUMERICAL,
    strlcpy(cap.card, cam->vdev->name, sizeof(cap.card));
    if (usb_make_path(cam->udev, cap.bus_info, sizeof(cap.bus_info)) < 0)
      strlcpy(cap.bus_info, cam->vdev->name, sizeof(cap.bus_info));

    if (copy_to_user(arg, &cap, sizeof(cap)))
      return -EFAULT;
    break;
  }

  /* --- capture ioctls ------------------------------------------ */
  case VIDIOC_ENUM_FMT:
  {
    struct v4l2_fmtdesc fmtd;

    if (copy_from_user(&fmtd, arg, sizeof(fmtd)))
      return -EFAULT;

    PDEBUG(DBG_TRACE, "VIDIOC_ENUM_FMT %d", fmtd.index);
    if (fmtd.index != 0)
      return -EINVAL;

    memset(&fmtd, 0, sizeof(fmtd));
    fmtd.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    fmtd.flags = 0;
    strcpy(fmtd.description, "rgb24");
    fmtd.pixelformat = V4L2_PIX_FMT_RGB24;
    memset(fmtd.reserved, 0, sizeof(fmtd.reserved));
    if (copy_to_user(arg, &fmtd, sizeof(fmtd)))
      return -EFAULT;
    break;
  }

  case VIDIOC_G_FMT:
  {
    struct v4l2_format fmtd ;
    struct v4l2_pix_format pix_format;

    if (copy_from_user(&fmtd, arg, sizeof(fmtd)))
      return -EFAULT;

    PDEBUG(DBG_TRACE, "GET FMT %d", fmtd.type);
    if (fmtd.type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
      retval = -EINVAL;

    pix_format.width = 640;
    pix_format.height = 480;
    pix_format.pixelformat = V4L2_PIX_FMT_RGB24;
    pix_format.field = V4L2_FIELD_NONE;
    pix_format.bytesperline = 3*pix_format.width;
    pix_format.sizeimage = pix_format.width * pix_format.height*3;
    pix_format.colorspace = V4L2_COLORSPACE_SRGB;
    pix_format.priv = 0;

    memcpy(&(fmtd.fmt.pix), &pix_format, sizeof(pix_format));
    if (copy_to_user(arg, &fmtd, sizeof(fmtd)))
      return -EFAULT;

    break;
  }

  case VIDIOC_S_FMT:
  {
    struct v4l2_format fmtd ;
    struct v4l2_pix_format pix_format;

    if (copy_from_user(&fmtd, arg, sizeof(fmtd)))
      return -EFAULT;

    PDEBUG(DBG_TRACE, "SET FMT %d", fmtd.type);
    PDEBUG(DBG_TRACE, "SET FMT width %d", fmtd.fmt.pix.width);
    PDEBUG(DBG_TRACE, "SET FMT height %d", fmtd.fmt.pix.height);
    if (fmtd.type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
      return -EINVAL;
    pix_format.width = 640;
    pix_format.height = 480;
    pix_format.pixelformat = V4L2_PIX_FMT_RGB24;
    pix_format.field = V4L2_FIELD_NONE;
    pix_format.bytesperline = 3*pix_format.width;
    pix_format.sizeimage = pix_format.width * pix_format.height*3;
    pix_format.colorspace = V4L2_COLORSPACE_SRGB;
    pix_format.priv = 0;

    memcpy(&(fmtd.fmt.pix), &pix_format, sizeof(pix_format));
    if (copy_to_user(arg, &fmtd, sizeof(fmtd)))
      return -EFAULT;


    break;
  }

  case VIDIOC_TRY_FMT:
  {
    struct v4l2_format fmtd ;
    struct v4l2_pix_format pix_format;

    if (copy_from_user(&fmtd, arg, sizeof(fmtd)))
      return -EFAULT;

    PDEBUG(DBG_TRACE, "TRY FMT %d", fmtd.type);
    if (fmtd.type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
      return -EINVAL;
    pix_format.width = 640;
    pix_format.height = 480;
    pix_format.pixelformat = V4L2_PIX_FMT_RGB24;
    pix_format.field = V4L2_FIELD_NONE;
    pix_format.bytesperline = 3*pix_format.width;
    pix_format.sizeimage = pix_format.width * pix_format.height;
    pix_format.colorspace = V4L2_COLORSPACE_SRGB;
    pix_format.priv = 0;

    memcpy(&(fmtd.fmt.pix), &pix_format, sizeof(pix_format));
    if (copy_to_user(arg, &fmtd, sizeof(fmtd)))
      return -EFAULT;

    break;
  }

  /* ------- tv norms --------- */
  case VIDIOC_ENUMSTD:
  {
    struct v4l2_standard e ;

    if (copy_from_user(&e, arg, sizeof(e)))
      return -EFAULT;

    PDEBUG(DBG_TRACE, "VIDIOC_ENUMSTD %d", e.index);
    if (e.index != 0)
      retval = -EINVAL;
    e.id = V4L2_STD_UNKNOWN;
//    e->id = V4L2_STD_PAL_BG;
//      e->id = V4L2_STD_PAL_G;
    if (copy_to_user( arg, &e, sizeof(e)))
      return -EFAULT;
    break;
  }

  case VIDIOC_G_STD:
  {
    struct v4l2_standard e ;

    if (copy_from_user(&e, arg, sizeof(e)))
      return -EFAULT;

    e.index = 0;
    PDEBUG(DBG_TRACE, "GET STD %d", e.index);
    if (copy_to_user(arg, &e,  sizeof(e)))
      return -EFAULT;
//       retval = -EINVAL;
    break;
  }

  case VIDIOC_S_STD:
  {
    struct v4l2_standard e ;

    if (copy_from_user(&e, arg, sizeof(e)))
      return -EFAULT;

    PDEBUG(DBG_TRACE, "SET STD %d", e.index);

    e.index = 0;
    if (copy_to_user(arg, &e,  sizeof(e)))
      return -EFAULT;
    break;
  }

  case VIDIOC_QUERYSTD:
  {
    struct v4l2_standard e ;

    if (copy_from_user(&e, arg, sizeof(e)))
      return -EFAULT;

    PDEBUG(DBG_TRACE, "QUERYSTD %d", e.index);

    e.index = 0;
    if (copy_to_user(arg, &e, sizeof(e)))
      return -EFAULT;
    break;
  }

  /* ------- input switching --------- */
  case VIDIOC_ENUMINPUT:
  {
    struct v4l2_input i ;

    if (copy_from_user(&i, arg, sizeof(i)))
      return -EFAULT;

    PDEBUG(DBG_TRACE, "VIDIOC_ENUMINPUT %d", i.index);
    strlcpy(i.name, "USB", sizeof(i.name));
    i.type = V4L2_INPUT_TYPE_CAMERA;
    if (copy_to_user(arg, &i, sizeof(i)))
      return -EFAULT;


    break;
  }

  case VIDIOC_G_INPUT:
  {
    struct v4l2_input i ;

    if (copy_from_user(&i, arg, sizeof(i)))
      return -EFAULT;
    PDEBUG(DBG_TRACE, "GET INPUT %d", i.index);
    if (i.index != 0)
      retval = -EINVAL;
    if (copy_to_user(arg, &i, sizeof(i)))
      return -EFAULT;
    break;
  }
  case VIDIOC_S_INPUT:
  {
    struct v4l2_input i ;

    if (copy_from_user(&i, arg, sizeof(i)))
      return -EFAULT;
    PDEBUG(DBG_TRACE, "SET INPUT %d", i.index);
    if (i.index != 0)
      retval = -EINVAL;
    if (copy_to_user(arg, &i, sizeof(i)))
      return -EFAULT;
    break;
  }

  case VIDIOC_REQBUFS:
  {
    struct v4l2_requestbuffers  rb ;
    u32 i;
    int err;

    if (copy_from_user(&rb, arg, sizeof(rb)))
      return -EFAULT;

    PDEBUG(DBG_TRACE, "VIDIOC_REQBUFS %d %d %d ",
                                              rb.count, rb.type, rb.memory);
    PDEBUG(DBG_TRACE, "VIDIOC_REQBUFS cam->io %d", cam->io);

    if (rb.type != V4L2_BUF_TYPE_VIDEO_CAPTURE ||
        rb.memory != V4L2_MEMORY_MMAP)
      return -EINVAL;

    if (cam->io == IO_READ) {
      PDEBUG(DBG_TRACE, "Close and open the device again to choose "
                                "the mmap I/O method");
      return -EINVAL;
    }

    for (i = 0; i < cam->nbuffers; i++) {
      if (cam->frame[i].vma_use_count) {
        PDEBUG(DBG_TRACE, "VIDIOC_REQBUFS failed. "
                                "Previous buffers are still mapped.");
        return -EINVAL;
      }
    }

    if (cam->stream == STREAM_ON)
      if ((err = m560x_stream_interrupt(cam)))
        return err;

    m560x_empty_framequeues(cam);

    m560x_release_buffers(cam);
    if (rb.count)
      rb.count = m560x_request_buffers(cam, rb.count,
                                         IO_MMAP);

    if (copy_to_user(arg, &rb, sizeof(rb))) {
       m560x_release_buffers(cam);
       cam->io = IO_NONE;
       return -EFAULT;
    }

    cam->io = rb.count ? IO_MMAP : IO_NONE;
    PDEBUG(DBG_TRACE, "VIDIOC_REQBUFS cam->io %d", cam->io);
    return 0;
  }

  case VIDIOC_QUERYBUF:
  {
    struct v4l2_buffer b ;

    if (copy_from_user(&b, arg, sizeof(b)))
      return -EFAULT;

    PDEBUG(DBG_TRACE, "VIDIOC_QUERYBUF %d %d %d",
                                        b.index, cam->nbuffers, cam->io);
    if (b.type != V4L2_BUF_TYPE_VIDEO_CAPTURE ||
        b.index >= cam->nbuffers || cam->io != IO_MMAP)
      return -EINVAL;

    PDEBUG(DBG_TRACE, "buffer length %08x",
                        (unsigned int)cam->frame[b.index].buf.length);
    memcpy(&b, &cam->frame[b.index].buf, sizeof(struct v4l2_buffer));

    if (cam->frame[b.index].vma_use_count)
      b.flags |= V4L2_BUF_FLAG_MAPPED;

    if (cam->frame[b.index].state == FRAME_DONE)
      b.flags |= V4L2_BUF_FLAG_DONE;
    else if (cam->frame[b.index].state != FRAME_UNUSED)
      b.flags |= V4L2_BUF_FLAG_QUEUED;

    if (copy_to_user(arg, &b, sizeof(b)))
      return -EFAULT;

    return 0;
  }

  case VIDIOC_QBUF:
  {
    struct v4l2_buffer  b ;
    unsigned long lock_flags;

    if (copy_from_user(&b, arg, sizeof(b)))
      return -EFAULT;

    PDEBUG(DBG_TRACE, "VIDIOC_QBUF");
    if (b.type != V4L2_BUF_TYPE_VIDEO_CAPTURE ||
        b.index >= cam->nbuffers || cam->io != IO_MMAP)
      return -EINVAL;

    if (cam->frame[b.index].state != FRAME_UNUSED)
      return -EINVAL;

    cam->frame[b.index].state = FRAME_QUEUED;

    spin_lock_irqsave(&cam->queue_lock, lock_flags);
    list_add_tail(&cam->frame[b.index].frame, &cam->inqueue);
    spin_unlock_irqrestore(&cam->queue_lock, lock_flags);

    PDEBUG(DBG_TRACE, "Frame #%lu queued", (unsigned long)b.index);

    return 0;
  }

  case VIDIOC_DQBUF:
  {
    struct v4l2_buffer b ;
    struct m560x_frame_t *f;
    unsigned long lock_flags;
    int err = 0;

    if (copy_from_user(&b, arg, sizeof(b)))
      return -EFAULT;


    PDEBUG(DBG_TRACE, "VIDIOC_DQBUF");
    if (b.type != V4L2_BUF_TYPE_VIDEO_CAPTURE || cam->io!= IO_MMAP)
      return -EINVAL;

    if (list_empty(&cam->outqueue)) {
      if (cam->stream == STREAM_OFF)
        return -EINVAL;
      if (file->f_flags & O_NONBLOCK)
        return -EAGAIN;
      err = wait_event_interruptible(cam->wait_frame,
                        (!list_empty(&cam->outqueue)) ||
                        (cam->state & (DEV_DISCONNECTED|DEV_MISCONFIGURED)) );
      if (err)
        return err;
      if (cam->state & DEV_DISCONNECTED)
        return -ENODEV;
      if (cam->state & DEV_MISCONFIGURED)
        return -EIO;
    }

    spin_lock_irqsave(&cam->queue_lock, lock_flags);
    f = list_entry(cam->outqueue.next, struct m560x_frame_t,
                   frame);
    list_del(cam->outqueue.next);
    spin_unlock_irqrestore(&cam->queue_lock, lock_flags);

    f->state = FRAME_UNUSED;
    memcpy(&b, &f->buf, sizeof(b));
    if (f->vma_use_count)
      b.flags |= V4L2_BUF_FLAG_MAPPED;

    if (copy_to_user(arg, &b, sizeof(b)))
      return -EFAULT;

    PDEBUG(DBG_TRACE, "Frame #%lu dequeued", (unsigned long)f->buf.index);

    return 0;
  }

  case VIDIOC_STREAMON:
  {
    int type;

    if (copy_from_user(&type, arg, sizeof(type)))
      return -EFAULT;

    if (type != V4L2_BUF_TYPE_VIDEO_CAPTURE || cam->io != IO_MMAP)
      return -EINVAL;

    if (list_empty(&cam->inqueue))
      return -EINVAL;

    cam->stream = STREAM_ON;

    PDEBUG(DBG_TRACE, "Stream on");

    return 0;
  }

  case VIDIOC_STREAMOFF:
  {
    int type;
    int err;

    if (copy_from_user(&type, arg, sizeof(type)))
      return -EFAULT;

    if (type != V4L2_BUF_TYPE_VIDEO_CAPTURE || cam->io != IO_MMAP)
      return -EINVAL;

    if (cam->stream == STREAM_ON)
      if ((err = m560x_stream_interrupt(cam)))
        return err;

    m560x_empty_framequeues(cam);

    PDEBUG(DBG_TRACE, "Stream off");

    return 0;
  }

  /* ------- streaming parameters --------- */
  case VIDIOC_G_PARM:
  {
    struct v4l2_streamparm sp;

    if (copy_from_user(&sp, arg, sizeof(sp)))
      return -EFAULT;


    PDEBUG(DBG_TRACE, "VIDIOC_G_PARM %d", sp.type);
    if (sp.type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
      retval = -EINVAL;

    sp.parm.capture.capability = 0;
    sp.parm.capture.capturemode = 0;
    sp.parm.capture.timeperframe.numerator = 1;
    sp.parm.capture.timeperframe.denominator = 30;
    sp.parm.capture.readbuffers = 2;
    sp.parm.capture.extendedmode = 0;

    if (copy_to_user(arg, &sp, sizeof(sp)))
      return -EFAULT;
    break;
  }

  /* audio interface - we have none */
  case VIDIOC_G_AUDIO:
  {
    PDEBUG(DBG_TRACE, "Get Audio");
    retval = -EINVAL;
    break;
  }
  case VIDIOC_S_AUDIO:
    retval = -EINVAL;
    break;

  /* ------- input switching --------- */
 // There are no parameters yet to control
  case VIDIOC_QUERYCTRL:
  {
    struct m560x_sensor *s = cam->sensor;
    struct v4l2_queryctrl qc;
    u8 i, n;

    if (copy_from_user(&qc, arg, sizeof(qc)))
      return -EFAULT;

    n = sizeof(s->qctrl) / sizeof(s->qctrl[0]);
    for (i = 0; i < n; i++)
      if (qc.id && qc.id == s->qctrl[i].id) {
      memcpy(&qc, &(s->qctrl[i]), sizeof(qc));
      if (copy_to_user(arg, &qc, sizeof(qc)))
        return -EFAULT;
      return 0;
      }

      return -EINVAL;
  }

  case VIDIOC_G_CTRL:
  {
    struct m560x_sensor *s = cam->sensor;
    struct v4l2_control ctrl;
    int err = 0;

    if (!s->get_ctrl)
      return -EINVAL;

    if (copy_from_user(&ctrl, arg, sizeof(ctrl)))
      return -EFAULT;

    err = s->get_ctrl(cam, &ctrl);
    PDEBUG(DBG_TRACE, "get_ctrl %d %d", err, ctrl.value);

    if (copy_to_user(arg, &ctrl, sizeof(ctrl)))
      return -EFAULT;

    return err;
  }

  case VIDIOC_S_CTRL_OLD:
  case VIDIOC_S_CTRL:
  {
    struct m560x_sensor *s = cam->sensor;
    struct v4l2_control ctrl;
    u8 i, n;
    int err = 0;

    PDEBUG(DBG_TRACE, "set_ctrl %d %d", err, ctrl.value);
    if (!s->set_ctrl)
      return -EINVAL;

    if (copy_from_user(&ctrl, arg, sizeof(ctrl)))
      return -EFAULT;

    n = sizeof(s->qctrl) / sizeof(s->qctrl[0]);
    for (i = 0; i < n; i++) {
      if (ctrl.id == s->qctrl[i].id) {
      if (ctrl.value < s->qctrl[i].minimum ||
          ctrl.value > s->qctrl[i].maximum)
        return -ERANGE;
      ctrl.value -= ctrl.value % s->qctrl[i].step;
      break;
      }
    }

    if ((err = s->set_ctrl(cam, &ctrl))) {
      PDEBUG(DBG_TRACE, "set_ctrl %d %d", err, ctrl.value);
      return err;
    }

    s->_qctrl[i].default_value = ctrl.value;

    PDEBUG(DBG_TRACE, "VIDIOC_S_CTRL: id %lu, value %lu",
                        (unsigned long)ctrl.id, (unsigned long)ctrl.value);

    return 0;
  }


  /* pointless to implement overlay with this camera */
  case VIDIOC_G_FBUF:
  case VIDIOC_S_FBUF:
  case VIDIOC_OVERLAY:
    retval = -EINVAL;
    break;

  /* tuner interface - we have none */
  case VIDIOC_S_TUNER:
  case VIDIOC_G_TUNER:
  case VIDIOC_G_FREQUENCY:
  case VIDIOC_S_FREQUENCY:
    retval = -EINVAL;
    break;



  case VIDIOC_QUERYMENU:
    retval = -EINVAL;
    break;

  case VIDIOC_CROPCAP:
  {
    struct v4l2_cropcap cc ;


    cc.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    cc.pixelaspect.numerator = 1;
    cc.pixelaspect.denominator = 1;
    cc.bounds.top = 0;
    cc.bounds.left = 0;
    cc.bounds.width = 640;
    cc.bounds.height = 480;
    cc.defrect.top = 0;
    cc.defrect.left = 0;
    cc.defrect.width = 640;
    cc.defrect.height = 480;

    if (copy_to_user(arg, &cc, sizeof(cc)))
      return -EFAULT;

    return 0;
  }

  case VIDIOC_G_CROP:
  case VIDIOC_S_CROP:
  {
    struct v4l2_crop crop;

    crop.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

    crop.c.top = 0;
    crop.c.left = 0;
    crop.c.width = 640;
    crop.c.height = 480;

    if (copy_to_user(arg, &crop, sizeof(crop)))
      return -EFAULT;

    return 0;
  }


  default:
   // return v4l_compat_translate_ioctl(inode, file, cmd, arg,
   //                                   video_do_ioctl);
 /*    PDEBUG(DBG_TRACE, "translate %02x", (unsigned char) ioctlnr);
   retval =  v4l_compat_translate_ioctl(inode, file, ioctlnr, (void __user *)arg, (void *) v4l_m560x_do_ioctl);
    PDEBUG(DBG_TRACE, "retval %08x", retval);*/
//       PDEBUG(DBG_TRACE, "invalid v4l_m560x_ioctl %02x", (unsigned char) ioctlnr);
    retval = -EINVAL;
    break;
  }

exit:
    return retval;
}



static int v4l_m560x_ioctl(struct inode *inode, struct file *file,
                           unsigned int cmd, unsigned long arg)
{
  struct m560x_camera * cam = video_get_drvdata(video_devdata(file));
  int err = 0;

  if (down_interruptible(&cam->fileop_sem))
    return -ERESTARTSYS;

  if (cam->state & DEV_DISCONNECTED) {
    PDEBUG(DBG_TRACE, "Device not present");
    err = -ENODEV;
    goto out;
  }

  if (cam->state & DEV_MISCONFIGURED) {
    PDEBUG(DBG_TRACE, "The camera is misconfigured. Close and open it again.");
    err = -EIO;
    goto out;
  }

  err = v4l_m560x_do_ioctl(inode, file,  cmd, (void __user *) arg);

out:
  up(&cam->fileop_sem);

  return err;
}


static struct file_operations v4l_m560x_fops = {
        .owner          = THIS_MODULE,
        .open           = v4l_m560x_open,
        .release        = v4l_m560x_release,
        .read           = v4l_m560x_read,
  .poll = v4l_m560x_poll,
        .mmap           = v4l_m560x_mmap,
        .ioctl          = v4l_m560x_ioctl,
 .llseek                = no_llseek,
/*#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 11)
     .compat_ioctl   = v4l_compat_ioctl32,
#endif*/
};



/****************************************************************************
 * USB - Functions
 ***************************************************************************/
static int usb_m560x_probe(struct usb_interface *intf,
                           const struct usb_device_id *id)
{
  struct usb_device *udev = interface_to_usbdev(intf);
  struct m560x_camera *cam = NULL;
  struct usb_host_interface *iface_desc;
  struct usb_endpoint_descriptor *endpoint;
  char *camera_name = NULL;
  size_t buffer_size;
  int i;
  int error = 0;

  info("m560x_probe - M560x-based WebCam connected");

  if (!(cam = kmalloc(sizeof(struct m560x_camera), GFP_KERNEL)))
    return -ENOMEM;
  memset(cam, 0, sizeof(struct m560x_camera));
  PDEBUG(DBG_TRACE, "cam %p", cam);
  cam->udev = udev;
  memcpy(&cam->dev, &udev->dev, sizeof(struct device));

  cam->name = camera_name;
  cam->camera_type = id->driver_info;
  cam->interface = intf;

  if (!(cam->control_buffer = kmalloc(0x100, GFP_KERNEL))) {
    err("kmalloc() failed");
    error = -ENOMEM;
    goto fail;
  }
  memset(cam->control_buffer, 0, 0x100);

  if (!(cam->hashtab = kmalloc(HASHSIZE*sizeof(struct nlist *), GFP_KERNEL))) {
    err("kmalloc() failed");
    error = -ENOMEM;
    goto fail;
  }
  memset(cam->hashtab, 0, HASHSIZE*sizeof(struct nlist *));

  /* set up the endpoint information */
  /* use only the first relevant endpoints */
  iface_desc = intf->cur_altsetting;

  for (i = 0; i < iface_desc->desc.bNumEndpoints; ++i) {
    endpoint = &iface_desc->endpoint[i].desc;

    if (!cam->int_in_endpointAddr
        && ((endpoint->bEndpointAddress & USB_ENDPOINT_DIR_MASK)
                                == USB_DIR_IN)
        && ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
                                == USB_ENDPOINT_XFER_INT)) {
        /* we found an int in endpoint */
      buffer_size = le16_to_cpu(endpoint->wMaxPacketSize);
      cam->int_in_size = buffer_size;
      cam->int_in_endpointAddr = endpoint->bEndpointAddress;
    }

    if (!cam->isoc_in_endpointAddr
        && ((endpoint->bEndpointAddress & USB_ENDPOINT_DIR_MASK)
                                == USB_DIR_IN)
        && ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
                                == USB_ENDPOINT_XFER_ISOC)) {
        /* we found an isoc in endpoint */
      buffer_size = le16_to_cpu(endpoint->wMaxPacketSize);
      cam->isoc_in_size = buffer_size;
      cam->isoc_in_endpointAddr = endpoint->bEndpointAddress;
        }
  }
  if (!(cam->int_in_endpointAddr && cam->isoc_in_endpointAddr)) {
    err("Could not find both int-in and isoc-in endpoints");
    goto fail;
  }

  cam->vdev = video_device_alloc();
  if (!cam->vdev) {
    error = -ENOMEM;
    goto fail;
  }

  init_MUTEX(&cam->dev_sem);


  if (initialize_camera(cam)) {
    PDEBUG(DBG_TRACE, "Initialization failed. I will retry on open().");
    cam->state |= DEV_MISCONFIGURED;
  }

  if (cam->camera_type == CAMERA_M5603) {
    for  (i = 0; m5603_sensor_table[i]; i++) {
      error = m5603_sensor_table[i](cam);
      if (!error)
        break;
    }
  } else if (cam->camera_type == CAMERA_M5602) {
    for  (i = 0; m5602_sensor_table[i]; i++) {
      error = m5602_sensor_table[i](cam);
      if (!error)
        break;
    }
  }

  if (!cam->sensor) {
    error = -ENOTSUPP;
    err("No supported sensor detected");
    goto fail;
  }

  strcpy(cam->vdev->name, "Ali M560x-based USB Camera");
  cam->vdev->owner    = THIS_MODULE;
  cam->vdev->type     = VID_TYPE_CAPTURE;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
  cam->vdev->hardware = VID_HARDWARE_M560X;
#endif
  cam->vdev->fops     = &v4l_m560x_fops;
  cam->vdev->release  = video_device_release;
  cam->vdev->minor    = -1;
  video_set_drvdata(cam->vdev, cam);
//   cam->users = 0;

  down(&cam->dev_sem);

  error = video_register_device(cam->vdev, VFL_TYPE_GRABBER, -1);

  if (error) {
    err("video_register_device failed");
    goto fail;
  }
  initialize_v4l_camera(cam);

  info("m560x_probe - Ali_M560x WebCam driver is now controlling "
                                "video device %d", cam->vdev->minor);

  m560x_create_sysfs_files(cam->vdev);

  usb_set_intfdata(intf, cam);

  up(&cam->dev_sem);

  return 0;

fail:
  if (cam) {
    if (cam->vdev)
      video_device_release(cam->vdev);
    if (cam->hashtab){
      clear_hashtab(cam->hashtab);
      kfree(cam->hashtab);
    }
    kfree(cam->control_buffer);
    kfree(cam);
  }
  return error;
}

static void usb_m560x_disconnect(struct usb_interface *intf)
{
  struct m560x_camera *cam = usb_get_intfdata(intf);

  PDEBUG(DBG_TRACE, "usb_m560x_disconnect");
  PDEBUG(DBG_TRACE, "cam-users %d", cam->users);
  if (!cam)
    return;

  down_write(&m560x_disconnect);

  down(&cam->dev_sem);

  wake_up_interruptible_all(&cam->open);

  if (cam->users) {
    PDEBUG(DBG_TRACE, "Device /dev/video%d is open! "
                        "Deregistration and memory deallocation are "
                        "deferred on close", cam->vdev->minor);
    cam->state |= DEV_MISCONFIGURED;
    m560x_stop_transfer(cam);
    cam->state |= DEV_DISCONNECTED;
    wake_up_interruptible(&cam->wait_frame);
    wake_up_interruptible(&cam->wait_stream);
  } else {
    cam->state |= DEV_DISCONNECTED;
    m560x_release_resources(cam);
  }

  up(&cam->dev_sem);

  if (!cam->users) {
    kfree(cam);
    PDEBUG(DBG_TRACE, "kfree cam");
  }

  up_write(&m560x_disconnect);

  info("m560x_disconnect- M560x-based WebCam disconnected");;
}

static struct usb_driver usb_m560x_driver = {
        .name = "m560x",
        .probe = usb_m560x_probe,
        .disconnect = usb_m560x_disconnect,
        .id_table = m560x_table,
};

static int __init usb_m560x_init(void)
{
	int result;

	info("usb_m560x_init - M560x-based WebCam driver startup");

	/* register this driver with the USB subsystem */
	result = usb_register(&usb_m560x_driver);
	if (result)
		err("usb_register failed. Error number %d", result);

	info(DRIVER_VERSION ":" DRIVER_DESC);

	return result;
}

static void __exit usb_m560x_exit(void)
{
	info("usb_m560x_exit - M560x-based WebCam driver shutdown");

	/* deregister this driver with the USB subsystem */
	usb_deregister(&usb_m560x_driver);
}


module_init (usb_m560x_init);
module_exit (usb_m560x_exit);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");
module_param(debug, int, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC (debug, "Debug enabled or not");
