/***************************************************************************
 * API for image sensors connected to the SN9C10x PC Camera Controllers    *
 *                                                                         *
 * Copyright (C) 2004-2005 by Luca Risolia <luca.risolia@studio.unibo.it>  *
 *                                                                         *
 * This program is free software; you can redistribute it and/or modify    *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation; either version 2 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * This program is distributed in the hope that it will be useful,         *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with this program; if not, write to the Free Software             *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               *
 ***************************************************************************/

#ifndef _M560X_SENSOR_H_
#define _M560X_SENSOR_H_

#include <linux/usb.h>
#include <linux/videodev.h>
#include <linux/device.h>
#include <linux/stddef.h>
#include <linux/errno.h>
#include <asm/types.h>

#define M560X_MAX_CTRLS V4L2_CID_LASTP1-V4L2_CID_BASE+10

struct m560x_camera;
struct m560x_sensor;

/*****************************************************************************/

/*
   OVERVIEW.
   This is a small interface that allows you to add support for any CCD/CMOS
   image sensors connected to the SN9C10X bridges. The entire API is documented
   below. In the most general case, to support a sensor there are three steps
   you have to follow:
   1) define the main "sn9c102_sensor" structure by setting the basic fields;
   2) write a probing function to be called by the core module when the USB
      camera is recognized, then add both the USB ids and the name of that
      function to the two corresponding tables SENSOR_TABLE and ID_TABLE (see
      below);
   3) implement the methods that you want/need (and fill the rest of the main
      structure accordingly).
   "sn9c102_pas106b.c" is an example of all this stuff. Remember that you do
   NOT need to touch the source code of the core module for the things to work
   properly, unless you find bugs or flaws in it. Finally, do not forget to
   read the V4L2 API for completeness.
*/

/*****************************************************************************/

/*
   Probing functions: on success, you must attach the sensor to the camera
   by calling sn9c102_attach_sensor() provided below.
   To enable the I2C communication, you might need to perform a really basic
   initialization of the SN9C10X chip by using the write function declared 
   ahead.
   Functions must return 0 on success, the appropriate error otherwise.
*/
extern int m5603_probe_mt9v011(struct m560x_camera* cam);
extern int m5603_probe_ov7640(struct m560x_camera* cam);
extern int m5602_probe_mt9v111(struct m560x_camera* cam);
extern int m5602_probe_ov9650(struct m560x_camera* cam);
extern int m5602_probe_ov7640(struct m560x_camera* cam);
/*
    Add the above entries to this table. Be sure to add the entry in the right
    place, since, on failure, the next probing routine is called according to 
    the order of the list below, from top to bottom.
*/

#define M5603_SENSOR_TABLE                                                  \
     static int (*m5603_sensor_table[])(struct m560x_camera*) = {           \
     &m5603_probe_mt9v011,                                                  \
     &m5603_probe_ov7640,                                                   \
     NULL,                                                                  \
};

#define M5602_SENSOR_TABLE                                                  \
     static int (*m5602_sensor_table[])(struct m560x_camera*) = {           \
     &m5602_probe_mt9v111,                                                  \
     &m5602_probe_ov9650,                                                   \
     &m5602_probe_ov7640,                                                   \
     NULL,                                                                  \
};
extern int m5603_init_from_script(struct m560x_camera *cam, char * pitem);
extern int read_sensor_register(struct m560x_camera *cam, __u16 ali_register, unsigned char * data);
extern int write_sensor_register(struct m560x_camera *cam, __u16 ali_register, unsigned char * data);
extern int read_m5602_register(struct m560x_camera *cam, unsigned short ali_register);
extern int write_m5602_register(struct m560x_camera *cam, unsigned short ali_register, unsigned char value);
  /* Attach a probed sensor to the camera. */
extern void 
    m560x_attach_sensor(struct m560x_camera* cam,
                          struct m560x_sensor* sensor);
struct m560x_sensor {
  char name[32], /* sensor name */
  maintainer[64]; /* name of the mantainer <email> */

  /* Supported operations through the 'sysfs' interface */
//   enum sn9c102_i2c_sysfs_ops sysfs_ops;

  /*
  These sensor capabilities must be provided if the SN9C10X controller
  needs to communicate through the sensor serial interface by using
  at least one of the i2c functions available.
  */
//   enum sn9c102_i2c_frequency frequency;
//   enum sn9c102_i2c_interface interface;

  /*
  This identifier must be provided if the image sensor implements
  the standard I2C protocol.
  */
  u8 i2c_slave_id; /* reg. 0x09 */

  /*
  NOTE: Where not noted,most of the functions below are not mandatory.
  Set to null if you do not implement them. If implemented,
  they must return 0 on success, the proper error otherwise.
  */

  int (*init)(struct m560x_camera* cam);
  /*
  This function will be called after the sensor has been attached. 
  It should be used to initialize the sensor only, but may also
  configure part of the SN9C10X chip if necessary. You don't need to
  setup picture settings like brightness, contrast, etc.. here, if
  the corrisponding controls are implemented (see below), since 
  they are adjusted in the core driver by calling the set_ctrl()
  method after init(), where the arguments are the default values
  specified in the v4l2_queryctrl list of supported controls;
  Same suggestions apply for other settings, _if_ the corresponding
  methods are present; if not, the initialization must configure the
  sensor according to the default configuration structures below.
  */

  struct v4l2_queryctrl qctrl[M560X_MAX_CTRLS];
  /*
  Optional list of default controls, defined as indicated in the 
  V4L2 API. Menu type controls are not handled by this interface.
  */

  int (*get_ctrl)(struct m560x_camera* cam, struct v4l2_control* ctrl);
  int (*set_ctrl)(struct m560x_camera* cam,
  const struct v4l2_control* ctrl);
  /*
  You must implement at least the set_ctrl method if you have defined
  the list above. The returned value must follow the V4L2
  specifications for the VIDIOC_G|C_CTRL ioctls. V4L2_CID_H|VCENTER
  are not supported by this driver, so do not implement them. Also,
  you don't have to check whether the passed values are out of bounds,
  given that this is done by the core module.
  */

  struct v4l2_cropcap cropcap;
  /*
  Think the image sensor as a grid of R,G,B monochromatic pixels
  disposed according to a particular Bayer pattern, which describes
  the complete array of pixels, from (0,0) to (xmax, ymax). We will
  use this coordinate system from now on. It is assumed the sensor
  chip can be programmed to capture/transmit a subsection of that
  array of pixels: we will call this subsection "active window".
  It is not always true that the largest achievable active window can
  cover the whole array of pixels. The V4L2 API defines another
  area called "source rectangle", which, in turn, is a subrectangle of
  the active window. The SN9C10X chip is always programmed to read the
  source rectangle.
  The bounds of both the active window and the source rectangle are
  specified in the cropcap substructures 'bounds' and 'defrect'.
  By default, the source rectangle should cover the largest possible
  area. Again, it is not always true that the largest source rectangle
  can cover the entire active window, although it is a rare case for 
  the hardware we have. The bounds of the source rectangle _must_ be
  multiple of 16 and must use the same coordinate system as indicated
  before; their centers shall align initially.
  If necessary, the sensor chip must be initialized during init() to
  set the bounds of the active sensor window; however, by default, it
  usually covers the largest achievable area (maxwidth x maxheight)
  of pixels, so no particular initialization is needed, if you have
  defined the correct default bounds in the structures.
  See the V4L2 API for further details.
  NOTE: once you have defined the bounds of the active window
  (struct cropcap.bounds) you must not change them.anymore.
  Only 'bounds' and 'defrect' fields are mandatory, other fields
  will be ignored.
  */

  int (*set_crop)(struct m560x_camera* cam,
  const struct v4l2_rect* rect);
  /*
  To be called on VIDIOC_C_SETCROP. The core module always calls a
  default routine which configures the appropriate SN9C10X regs (also
  scaling), but you may need to override/adjust specific stuff.
  'rect' contains width and height values that are multiple of 16: in
  case you override the default function, you always have to program
  the chip to match those values; on error return the corresponding
  error code without rolling back.
  NOTE: in case, you must program the SN9C10X chip to get rid of 
  blank pixels or blank lines at the _start_ of each line or
  frame after each HSYNC or VSYNC, so that the image starts with
  real RGB data (see regs 0x12, 0x13) (having set H_SIZE and,
  V_SIZE you don't have to care about blank pixels or blank
  lines at the end of each line or frame).
  */

  struct v4l2_pix_format pix_format;
  /*
  What you have to define here are: 1) initial 'width' and 'height' of
  the target rectangle 2) the initial 'pixelformat', which can be
  either V4L2_PIX_FMT_SN9C10X (for compressed video) or
  V4L2_PIX_FMT_SBGGR8 3) 'priv', which we'll be used to indicate the
  number of bits per pixel for uncompressed video, 8 or 9 (despite the
  current value of 'pixelformat').
  NOTE 1: both 'width' and 'height' _must_ be either 1/1 or 1/2 or 1/4
  of cropcap.defrect.width and cropcap.defrect.height. I
  suggest 1/1.
  NOTE 2: The initial compression quality is defined by the first bit
  of reg 0x17 during the initialization of the image sensor.
  NOTE 3: as said above, you have to program the SN9C10X chip to get
  rid of any blank pixels, so that the output of the sensor
  matches the RGB bayer sequence (i.e. BGBGBG...GRGRGR).
  */

  int (*set_pix_format)(struct m560x_camera* cam,
  const struct v4l2_pix_format* pix);
  /*
  To be called on VIDIOC_S_FMT, when switching from the SBGGR8 to
  SN9C10X pixel format or viceversa. On error return the corresponding
  error code without rolling back.
  */
  int (*i2c_write)(struct m560x_camera* cam, const unsigned char register_id, const unsigned char * data);
  int (*i2c_read)(struct m560x_camera* cam, const unsigned char register_id, unsigned char * data);
  const struct device* dev;
  /*
  This is the argument for dev_err(), dev_info() and dev_warn(). It
  is used for debugging purposes. You must not access the struct
  before the sensor is attached.
  */

  const struct usb_device* usbdev;
  /*
  Points to the usb_device struct after the sensor is attached.
  Do not touch unless you know what you are doing.
  */

  /*
  Do NOT write to the data below, it's READ ONLY. It is used by the
  core module to store successfully updated values of the above
  settings, for rollbacks..etc..in case of errors during atomic I/O
  */
  struct v4l2_queryctrl _qctrl[M560X_MAX_CTRLS];
  struct v4l2_rect _rect;
};

#define M560X_V4L2_CID_GREEN_BALANCE V4L2_CID_PRIVATE_BASE
#define M560X_V4L2_CID_EXTRA_GAIN V4L2_CID_PRIVATE_BASE+1

#endif /* _M560X_SENSOR_H_ */
